"""Generic class to handle Image-like data in a OME-NGFF file."""

import warnings
from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from typing import Any, Literal

import dask.array as da
import numpy as np
import zarr
from zarr.core.array import CompressorLike

from ngio.common import (
    Dimensions,
    InterpolationOrder,
    Roi,
    consolidate_pyramid,
)
from ngio.common._pyramid import ChunksLike, ShardsLike, shapes_from_scaling_factors
from ngio.images._create_utils import (
    _image_or_label_meta,
    init_image_like_from_shapes,
)
from ngio.io_pipes import (
    DaskGetter,
    DaskRoiGetter,
    DaskRoiSetter,
    DaskSetter,
    NumpyGetter,
    NumpyRoiGetter,
    NumpyRoiSetter,
    NumpySetter,
    SlicingInputType,
    TransformProtocol,
)
from ngio.ome_zarr_meta import (
    AxesHandler,
    Dataset,
    ImageMetaHandler,
    LabelMetaHandler,
    NgioImageMeta,
    PixelSize,
)
from ngio.ome_zarr_meta.ngio_specs import (
    Channel,
    NgffVersions,
    NgioLabelMeta,
)
from ngio.tables import RoiTable
from ngio.utils import (
    NgioFileExistsError,
    NgioValueError,
    StoreOrGroup,
    ZarrGroupHandler,
)
from ngio.utils._zarr_utils import find_dimension_separator


class AbstractImage(ABC):
    """A class to handle a single image (or level) in an OME-Zarr image.

    This class is meant to be subclassed by specific image types.
    """

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        path: str,
        meta_handler: ImageMetaHandler | LabelMetaHandler,
    ) -> None:
        """Initialize the Image at a single level.

        Args:
            group_handler: The Zarr group handler.
            path: The path to the image in the ome_zarr file.
            meta_handler: The image metadata handler.

        """
        self._path = path
        self._group_handler = group_handler
        self._meta_handler = meta_handler

        try:
            self._zarr_array = self._group_handler.get_array(self._path)
        except NgioFileExistsError as e:
            raise NgioFileExistsError(f"Could not find the dataset at {path}.") from e

    def __repr__(self) -> str:
        """Return a string representation of the image."""
        return f"Image(path={self.path}, {self.dimensions})"

    @property
    def path(self) -> str:
        """Return the path of the image."""
        return self._path

    @property
    @abstractmethod
    def meta_handler(self) -> ImageMetaHandler | LabelMetaHandler:
        """Return the metadata."""
        pass

    @property
    @abstractmethod
    def meta(self) -> NgioImageMeta | NgioLabelMeta:
        """Return the metadata."""
        pass

    @property
    def dataset(self) -> Dataset:
        """Return the dataset of the image."""
        return self.meta_handler.get_meta().get_dataset(path=self.path)

    @property
    def dimensions(self) -> Dimensions:
        """Return the dimensions of the image."""
        return Dimensions(
            shape=self.zarr_array.shape,
            chunks=self.zarr_array.chunks,
            dataset=self.dataset,
        )

    @property
    def pixel_size(self) -> PixelSize:
        """Return the pixel size of the image."""
        return self.dataset.pixel_size

    @property
    def axes_handler(self) -> AxesHandler:
        """Return the axes handler of the image."""
        return self.dataset.axes_handler

    @property
    def axes(self) -> tuple[str, ...]:
        """Return the axes of the image."""
        return self.dimensions.axes

    @property
    def zarr_array(self) -> zarr.Array:
        """Return the Zarr array."""
        return self._zarr_array

    @property
    def shape(self) -> tuple[int, ...]:
        """Return the shape of the image."""
        return self.zarr_array.shape

    @property
    def dtype(self) -> str:
        """Return the dtype of the image."""
        return str(self.zarr_array.dtype)

    @property
    def chunks(self) -> tuple[int, ...]:
        """Return the chunks of the image."""
        return self.zarr_array.chunks

    @property
    def is_3d(self) -> bool:
        """Return True if the image is 3D."""
        return self.dimensions.is_3d

    @property
    def is_2d(self) -> bool:
        """Return True if the image is 2D."""
        return self.dimensions.is_2d

    @property
    def is_time_series(self) -> bool:
        """Return True if the image is a time series."""
        return self.dimensions.is_time_series

    @property
    def is_2d_time_series(self) -> bool:
        """Return True if the image is a 2D time series."""
        return self.dimensions.is_2d_time_series

    @property
    def is_3d_time_series(self) -> bool:
        """Return True if the image is a 3D time series."""
        return self.dimensions.is_3d_time_series

    @property
    def is_multi_channels(self) -> bool:
        """Return True if the image is multichannel."""
        return self.dimensions.is_multi_channels

    @property
    def space_unit(self) -> str | None:
        """Return the space unit of the image."""
        return self.axes_handler.space_unit

    @property
    def time_unit(self) -> str | None:
        """Return the time unit of the image."""
        return self.axes_handler.time_unit

    def has_axis(self, axis: str) -> bool:
        """Return True if the image has the given axis."""
        return self.axes_handler.has_axis(axis)

    def _get_as_numpy(
        self,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: SlicingInputType,
    ) -> np.ndarray:
        """Get the image as a numpy array.

        Args:
            axes_order: The order of the axes to return the array.
            transforms: The transforms to apply to the array.
            **slicing_kwargs: The slices to get the array.

        Returns:
            The array of the region of interest.
        """
        numpy_getter = NumpyGetter(
            zarr_array=self.zarr_array,
            dimensions=self.dimensions,
            axes_order=axes_order,
            transforms=transforms,
            slicing_dict=slicing_kwargs,
        )
        return numpy_getter()

    def _get_roi_as_numpy(
        self,
        roi: Roi,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: SlicingInputType,
    ) -> np.ndarray:
        """Get the image as a numpy array for a region of interest.

        Args:
            roi: The region of interest to get the array.
            axes_order: The order of the axes to return the array.
            transforms: The transforms to apply to the array.
            **slicing_kwargs: The slices to get the array.

        Returns:
            The array of the region of interest.
        """
        numpy_roi_getter = NumpyRoiGetter(
            zarr_array=self.zarr_array,
            dimensions=self.dimensions,
            roi=roi,
            axes_order=axes_order,
            transforms=transforms,
            slicing_dict=slicing_kwargs,
        )
        return numpy_roi_getter()

    def _get_as_dask(
        self,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: SlicingInputType,
    ) -> da.Array:
        """Get the image as a dask array.

        Args:
            axes_order: The order of the axes to return the array.
            transforms: The transforms to apply to the array.
            **slicing_kwargs: The slices to get the array.
        """
        dask_getter = DaskGetter(
            zarr_array=self.zarr_array,
            dimensions=self.dimensions,
            axes_order=axes_order,
            transforms=transforms,
            slicing_dict=slicing_kwargs,
        )
        return dask_getter()

    def _get_roi_as_dask(
        self,
        roi: Roi,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: SlicingInputType,
    ) -> da.Array:
        """Get the image as a dask array for a region of interest.

        Args:
            roi: The region of interest to get the array.
            axes_order: The order of the axes to return the array.
            transforms: The transforms to apply to the array.
            **slicing_kwargs: The slices to get the array.
        """
        roi_dask_getter = DaskRoiGetter(
            zarr_array=self.zarr_array,
            dimensions=self.dimensions,
            roi=roi,
            axes_order=axes_order,
            transforms=transforms,
            slicing_dict=slicing_kwargs,
        )
        return roi_dask_getter()

    def _get_array(
        self,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        mode: Literal["numpy", "dask"] = "numpy",
        **slicing_kwargs: SlicingInputType,
    ) -> np.ndarray | da.Array:
        """Get a slice of the image.

        Args:
            axes_order: The order of the axes to return the array.
            transforms: The transforms to apply to the array.
            mode: The object type to return.
                Can be "dask", "numpy".
            **slicing_kwargs: The slices to get the array.

        Returns:
            The array of the region of interest.
        """
        if mode == "numpy":
            return self._get_as_numpy(
                axes_order=axes_order, transforms=transforms, **slicing_kwargs
            )
        elif mode == "dask":
            return self._get_as_dask(
                axes_order=axes_order, transforms=transforms, **slicing_kwargs
            )
        else:
            raise ValueError(
                f"Unsupported mode: {mode}. Supported modes are: numpy, dask."
            )

    def _get_roi(
        self,
        roi: Roi,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        mode: Literal["numpy", "dask"] = "numpy",
        **slice_kwargs: SlicingInputType,
    ) -> np.ndarray | da.Array:
        """Get a slice of the image.

        Args:
            roi: The region of interest to get the array.
            axes_order: The order of the axes to return the array.
            transforms: The transforms to apply to the array.
            mode: The mode to return the array.
                Can be "dask", "numpy".
            **slice_kwargs: The slices to get the array.

        Returns:
            The array of the region of interest.
        """
        if mode == "numpy":
            return self._get_roi_as_numpy(
                roi=roi, axes_order=axes_order, transforms=transforms, **slice_kwargs
            )
        elif mode == "dask":
            return self._get_roi_as_dask(
                roi=roi, axes_order=axes_order, transforms=transforms, **slice_kwargs
            )
        else:
            raise ValueError(
                f"Unsupported mode: {mode}. Supported modes are: numpy, dask."
            )

    def _set_array(
        self,
        patch: np.ndarray | da.Array,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: SlicingInputType,
    ) -> None:
        """Set a slice of the image.

        Args:
            patch: The patch to set.
            axes_order: The order of the axes to set the patch.
            transforms: The transforms to apply to the patch.
            **slicing_kwargs: The slices to set the patch.

        """
        if isinstance(patch, np.ndarray):
            numpy_setter = NumpySetter(
                zarr_array=self.zarr_array,
                dimensions=self.dimensions,
                axes_order=axes_order,
                transforms=transforms,
                slicing_dict=slicing_kwargs,
            )
            numpy_setter(patch)

        elif isinstance(patch, da.Array):
            dask_setter = DaskSetter(
                zarr_array=self.zarr_array,
                dimensions=self.dimensions,
                axes_order=axes_order,
                transforms=transforms,
                slicing_dict=slicing_kwargs,
            )
            dask_setter(patch)
        else:
            raise TypeError(
                f"Unsupported patch type: {type(patch)}. "
                "Supported types are: "
                "numpy.ndarray, dask.array.Array."
            )

    def _set_roi(
        self,
        roi: Roi,
        patch: np.ndarray | da.Array,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: SlicingInputType,
    ) -> None:
        """Set a slice of the image.

        Args:
            roi: The region of interest to set the patch.
            patch: The patch to set.
            axes_order: The order of the axes to set the patch.
            transforms: The transforms to apply to the patch.
            **slicing_kwargs: The slices to set the patch.

        """
        if isinstance(patch, np.ndarray):
            roi_numpy_setter = NumpyRoiSetter(
                zarr_array=self.zarr_array,
                dimensions=self.dimensions,
                roi=roi,
                axes_order=axes_order,
                transforms=transforms,
                slicing_dict=slicing_kwargs,
            )
            roi_numpy_setter(patch)

        elif isinstance(patch, da.Array):
            roi_dask_setter = DaskRoiSetter(
                zarr_array=self.zarr_array,
                dimensions=self.dimensions,
                roi=roi,
                axes_order=axes_order,
                transforms=transforms,
                slicing_dict=slicing_kwargs,
            )
            roi_dask_setter(patch)
        else:
            raise TypeError(
                f"Unsupported patch type: {type(patch)}. "
                "Supported types are: "
                "numpy.ndarray, dask.array.Array."
            )

    def _consolidate(
        self,
        order: InterpolationOrder = "linear",
        mode: Literal["dask", "numpy", "coarsen"] = "dask",
    ) -> None:
        """Consolidate the image on disk.

        Args:
            order: The order of the consolidation.
            mode: The mode of the consolidation.
        """
        consolidate_image(image=self, order=order, mode=mode)

    def roi(self, name: str | None = "image") -> Roi:
        """Return the ROI covering the entire image."""
        slices = {}
        for ax_name in ["t", "z", "y", "x"]:
            axis_size = self.dimensions.get(ax_name, default=None)
            if axis_size is None:
                continue
            slices[ax_name] = slice(0, axis_size)
        roi_px = Roi.from_values(name=name, slices=slices, space="pixel")
        return roi_px.to_world(pixel_size=self.pixel_size)

    def build_image_roi_table(self, name: str | None = "image") -> RoiTable:
        """Build the ROI table containing the ROI covering the entire image."""
        return RoiTable(rois=[self.roi(name=name)])

    def require_dimensions_match(
        self,
        other: "AbstractImage",
        allow_singleton: bool = False,
    ) -> None:
        """Assert that two images have matching spatial dimensions.

        Args:
            other: The other image to compare to.
            allow_singleton: If True, allow singleton dimensions to be
                compatible with non-singleton dimensions.

        Raises:
            NgioValueError: If the images do not have compatible dimensions.
        """
        self.dimensions.require_dimensions_match(
            other.dimensions, allow_singleton=allow_singleton
        )

    def check_if_dimensions_match(
        self,
        other: "AbstractImage",
        allow_singleton: bool = False,
    ) -> bool:
        """Check if two images have matching spatial dimensions.

        Args:
            other: The other image to compare to.
            allow_singleton: If True, allow singleton dimensions to be
                compatible with non-singleton dimensions.

        Returns:
            bool: True if the images have matching dimensions, False otherwise.
        """
        return self.dimensions.check_if_dimensions_match(
            other.dimensions, allow_singleton=allow_singleton
        )

    def require_axes_match(
        self,
        other: "AbstractImage",
    ) -> None:
        """Assert that two images have compatible axes.

        Args:
            other: The other image to compare to.

        Raises:
            NgioValueError: If the images do not have compatible axes.
        """
        self.dimensions.require_axes_match(other.dimensions)

    def check_if_axes_match(
        self,
        other: "AbstractImage",
    ) -> bool:
        """Check if two images have compatible axes.

        Args:
            other: The other image to compare to.

        Returns:
            bool: True if the images have compatible axes, False otherwise.

        """
        return self.dimensions.check_if_axes_match(other.dimensions)

    def require_rescalable(
        self,
        other: "AbstractImage",
    ) -> None:
        """Assert that two images can be rescaled to each other.

        For this to be true, the images must have the same axes, and
        the pixel sizes must be compatible (i.e. one can be scaled to the other).

        Args:
            other: The other image to compare to.

        Raises:
            NgioValueError: If the images cannot be scaled to each other.
        """
        self.dimensions.require_rescalable(other.dimensions)

    def check_if_rescalable(
        self,
        other: "AbstractImage",
    ) -> bool:
        """Check if two images can be rescaled to each other.

        For this to be true, the images must have the same axes, and
        the pixel sizes must be compatible (i.e. one can be scaled to the other).

        Args:
            other: The other image to compare to.

        Returns:
            bool: True if the images can be rescaled to each other, False otherwise.
        """
        return self.dimensions.check_if_rescalable(other.dimensions)


def consolidate_image(
    image: AbstractImage,
    order: InterpolationOrder = "linear",
    mode: Literal["dask", "numpy", "coarsen"] = "dask",
) -> None:
    """Consolidate the image on disk."""
    target_paths = image.meta_handler.get_meta().paths
    targets = [
        image._group_handler.get_array(path)
        for path in target_paths
        if path != image.path
    ]
    consolidate_pyramid(
        source=image.zarr_array, targets=targets, order=order, mode=mode
    )


def _shapes_from_ref_image(
    ref_image: AbstractImage,
) -> list[tuple[int, ...]]:
    """Rebuild base shape based on a new shape."""
    paths = ref_image.meta.paths
    index_path = paths.index(ref_image.path)
    sub_paths = paths[index_path:]
    group_handler = ref_image._group_handler
    shapes = []
    for path in sub_paths:
        zarr_array = group_handler.get_array(path)
        shapes.append(zarr_array.shape)
    if len(shapes) == len(paths):
        return shapes
    missing_levels = len(paths) - len(shapes)
    print(ref_image.meta.scaling_factor())
    extended_shapes = shapes_from_scaling_factors(
        base_shape=shapes[-1],
        scaling_factors=ref_image.meta.scaling_factor(),
        num_levels=missing_levels + 1,
    )
    shapes.extend(extended_shapes[1:])
    return shapes


def _shapes_from_new_shape(
    ref_image: AbstractImage,
    shape: Sequence[int],
) -> list[tuple[int, ...]]:
    """Rebuild pyramid shapes based on a new base shape."""
    if len(shape) != len(ref_image.shape):
        raise NgioValueError(
            "The shape of the new image does not match the reference image."
            f"Got shape {shape} for reference shape {ref_image.shape}."
        )
    base_shape = tuple(shape)
    scaling_factors = ref_image.meta.scaling_factor()
    num_levels = len(ref_image.meta.paths)
    return shapes_from_scaling_factors(
        base_shape=base_shape,
        scaling_factors=scaling_factors,
        num_levels=num_levels,
    )


def _compute_pyramid_shapes(
    ref_image: AbstractImage,
    shape: Sequence[int] | None,
) -> list[tuple[int, ...]]:
    """Rebuild pyramid shapes based on a new base shape."""
    if shape is None:
        return _shapes_from_ref_image(ref_image=ref_image)
    return _shapes_from_new_shape(ref_image=ref_image, shape=shape)


def _check_chunks_and_shards_compatibility(
    ref_shape: tuple[int, ...],
    chunks: ChunksLike,
    shards: ShardsLike | None,
) -> None:
    """Check if the chunks and shards are compatible with the reference shape.

    Args:
        ref_shape: The reference shape.
        chunks: The chunks to check.
        shards: The shards to check.
    """
    if chunks != "auto":
        if len(chunks) != len(ref_shape):
            raise NgioValueError(
                "The length of the chunks must be the same as the number of dimensions."
            )
    if shards is not None and shards != "auto":
        if len(shards) != len(ref_shape):
            raise NgioValueError(
                "The length of the shards must be the same as the number of dimensions."
            )


def _apply_channel_policy(
    ref_image: AbstractImage,
    channels_policy: Literal["squeeze", "same", "singleton"] | int,
    shapes: list[tuple[int, ...]],
    axes: tuple[str, ...],
    chunks: ChunksLike,
    shards: ShardsLike | None,
) -> tuple[list[tuple[int, ...]], tuple[str, ...], ChunksLike, ShardsLike | None]:
    """Apply the channel policy to the shapes and axes.

    Args:
        ref_image: The reference image.
        channels_policy: The channels policy to apply.
        shapes: The shapes of the pyramid levels.
        axes: The axes of the image.
        chunks: The chunks of the image.
        shards: The shards of the image.

    Returns:
        The new shapes and axes after applying the channel policy.
    """
    if channels_policy == "same":
        return shapes, axes, chunks, shards

    if channels_policy == "singleton":
        # Treat 'singleton' as setting channel size to 1
        channels_policy = 1

    channel_index = ref_image.axes_handler.get_index("c")
    if channel_index is None:
        if channels_policy == "squeeze":
            return shapes, axes, chunks, shards
        raise NgioValueError(
            f"Cannot apply channel policy {channels_policy=} to an image "
            "without channels axis."
        )
    if channels_policy == "squeeze":
        new_shapes = []
        for shape in shapes:
            new_shape = shape[:channel_index] + shape[channel_index + 1 :]
            new_shapes.append(new_shape)
        new_axes = axes[:channel_index] + axes[channel_index + 1 :]
        if chunks == "auto":
            new_chunks: ChunksLike = "auto"
        else:
            new_chunks = chunks[:channel_index] + chunks[channel_index + 1 :]
        if shards == "auto" or shards is None:
            new_shards: ShardsLike | None = shards
        else:
            new_shards = shards[:channel_index] + shards[channel_index + 1 :]
        return new_shapes, new_axes, new_chunks, new_shards
    elif isinstance(channels_policy, int):
        new_shapes = []
        for shape in shapes:
            new_shape = (
                *shape[:channel_index],
                channels_policy,
                *shape[channel_index + 1 :],
            )
            new_shapes.append(new_shape)
        return new_shapes, axes, chunks, shards
    else:
        raise NgioValueError(
            f"Invalid channels policy: {channels_policy}. "
            "Must be 'squeeze', 'same', or an integer."
        )


def _check_channels_meta_compatibility(
    meta_type: type[_image_or_label_meta],
    ref_image: AbstractImage,
    channels_meta: Sequence[str | Channel] | None,
) -> Sequence[str | Channel] | None:
    """Check if the channels metadata is compatible with the reference image.

    Args:
        meta_type: The metadata type.
        ref_image: The reference image.
        channels_meta: The channels metadata to check.

    Returns:
        The channels metadata if compatible, None otherwise.
    """
    if issubclass(meta_type, NgioLabelMeta):
        if channels_meta is not None:
            raise NgioValueError("Cannot set channels_meta for a label image.")
        return None
    if channels_meta is not None:
        return channels_meta
    assert isinstance(ref_image.meta, NgioImageMeta)
    ref_meta = ref_image.meta
    index_c = ref_meta.axes_handler.get_index("c")
    if index_c is None:
        return None

    # If the channels number does not match, return None
    # Else return the channels metadata from the reference image
    ref_shape = ref_image.shape
    ref_num_channels = ref_shape[index_c] if index_c is not None else 1
    channels_ = ref_meta.channels_meta.channels if ref_meta.channels_meta else []
    # Reset to None if number of channels do not match
    channels_meta_ = channels_ if ref_num_channels == len(channels_) else None
    return channels_meta_


def abstract_derive(
    *,
    ref_image: AbstractImage,
    meta_type: type[_image_or_label_meta],
    store: StoreOrGroup,
    overwrite: bool = False,
    # Metadata parameters
    shape: Sequence[int] | None = None,
    pixelsize: float | tuple[float, float] | None = None,
    z_spacing: float | None = None,
    time_spacing: float | None = None,
    name: str | None = None,
    channels_policy: Literal["squeeze", "same", "singleton"] | int = "same",
    channels_meta: Sequence[str | Channel] | None = None,
    ngff_version: NgffVersions | None = None,
    # Zarr Array parameters
    chunks: ChunksLike = "auto",
    shards: ShardsLike | None = None,
    dtype: str | None = None,
    dimension_separator: Literal[".", "/"] | None = None,
    compressors: CompressorLike | None = None,
    extra_array_kwargs: Mapping[str, Any] | None = None,
    # Deprecated arguments
    labels: Sequence[str] | None = None,
    pixel_size: PixelSize | None = None,
) -> ZarrGroupHandler:
    """Create an empty OME-Zarr image from an existing image.

    If a kwarg is not provided, the value from the reference image will be used.

    Args:
        ref_image (AbstractImage): The reference image to derive from.
        meta_type (type[_image_or_label_meta]): The metadata type to use.
        store (StoreOrGroup): The Zarr store or group to create the image in.
        overwrite (bool): Whether to overwrite an existing image.
        shape (Sequence[int] | None): The shape of the new image.
        pixelsize (float | tuple[float, float] | None): The pixel size of the new image.
        z_spacing (float | None): The z spacing of the new image.
        time_spacing (float | None): The time spacing of the new image.
        axes_names (Sequence[str] | None): The axes names of the new image.
        name (str | None): The name of the new image.
        channels_policy (Literal["squeeze", "same", "singleton"] | int):
            Possible policies:
            - If "squeeze", the channels axis will be removed (no matter its size).
            - If "same", the channels axis will be kept as is (if it exists).
            - If "singleton", the channels axis will be set to size 1.
            - If an integer is provided, the channels axis will be changed to have that
                size.
        channels_meta (Sequence[str | Channel] | None): The channels metadata
            of the new image.
        ngff_version (NgffVersions | None): The NGFF version to use.
        chunks (Sequence[int] | None): The chunk shape of the new image.
        shards (ShardsLike | None): The shard shape of the new image.
        dtype (str | None): The data type of the new image.
        dimension_separator (DIMENSION_SEPARATOR | None): The separator to use for
            dimensions.
        compressors (CompressorLike | None): The compressors to use.
        extra_array_kwargs (Mapping[str, Any] | None): Extra arguments to pass to
            the zarr array creation.
        labels (Sequence[str] | None): The labels of the new image.
            This argument is DEPRECATED please use channels_meta instead.
        pixel_size (PixelSize | None): The pixel size of the new image.
            This argument is DEPRECATED please use pixelsize, z_spacing,
            and time_spacing instead.

    Returns:
        ImagesContainer: The new derived image.

    """
    # TODO: remove in ngio 0.6
    if labels is not None:
        warnings.warn(
            "The 'labels' argument is deprecated and will be removed in "
            "a future release.",
            DeprecationWarning,
            stacklevel=2,
        )
        channels_meta = list(labels)
    if pixel_size is not None:
        warnings.warn(
            "The 'pixel_size' argument is deprecated and will be removed in "
            "a future release.",
            DeprecationWarning,
            stacklevel=2,
        )
        pixelsize_ = (pixel_size.y, pixel_size.x)
        z_spacing_ = pixel_size.z
        time_spacing_ = pixel_size.t
    else:
        if pixelsize is None:
            pixelsize_ = (ref_image.pixel_size.y, ref_image.pixel_size.x)
        else:
            pixelsize_ = pixelsize

        if z_spacing is None:
            z_spacing_ = ref_image.pixel_size.z
        else:
            z_spacing_ = z_spacing

        if time_spacing is None:
            time_spacing_ = ref_image.pixel_size.t
        else:
            time_spacing_ = time_spacing
    ref_meta = ref_image.meta

    shapes = _compute_pyramid_shapes(
        shape=shape,
        ref_image=ref_image,
    )
    ref_shape = next(iter(shapes))

    if pixelsize is None:
        pixelsize = (ref_image.pixel_size.y, ref_image.pixel_size.x)

    if z_spacing is None:
        z_spacing = ref_image.pixel_size.z

    if time_spacing is None:
        time_spacing = ref_image.pixel_size.t

    if name is None:
        name = ref_meta.name

    if dtype is None:
        dtype = ref_image.dtype

    if dimension_separator is None:
        dimension_separator = find_dimension_separator(ref_image.zarr_array)

    if compressors is None:
        compressors = ref_image.zarr_array.compressors  # type: ignore

    if chunks is None:
        chunks = ref_image.zarr_array.chunks

    if shards is None:
        shards = ref_image.zarr_array.shards

    _check_chunks_and_shards_compatibility(
        ref_shape=ref_shape,
        chunks=chunks,
        shards=shards,
    )

    if ngff_version is None:
        ngff_version = ref_meta.version

    shapes, axes, chunks, shards = _apply_channel_policy(
        ref_image=ref_image,
        channels_policy=channels_policy,
        shapes=shapes,
        axes=ref_image.axes,
        chunks=chunks,
        shards=shards,
    )
    channels_meta_ = _check_channels_meta_compatibility(
        meta_type=meta_type,
        ref_image=ref_image,
        channels_meta=channels_meta,
    )

    handler = init_image_like_from_shapes(
        store=store,
        meta_type=meta_type,
        shapes=shapes,
        pixelsize=pixelsize_,
        z_spacing=z_spacing_,
        time_spacing=time_spacing_,
        levels=ref_meta.paths,
        time_unit=ref_image.time_unit,
        space_unit=ref_image.space_unit,
        axes_names=axes,
        name=name,
        channels_meta=channels_meta_,
        chunks=chunks,
        shards=shards,
        dtype=dtype,
        dimension_separator=dimension_separator,
        compressors=compressors,
        overwrite=overwrite,
        ngff_version=ngff_version,
        extra_array_kwargs=extra_array_kwargs,
    )
    return handler



"""Implementation of a generic table class."""

import builtins
from abc import ABC, abstractmethod
from typing import Literal, Self

import pandas as pd
import polars as pl
from anndata import AnnData

from ngio.tables.backends import (
    BackendMeta,
    DefaultTableBackend,
    ImplementedTableBackends,
    TableBackend,
    TableBackendProtocol,
    TabularData,
    convert_to_anndata,
    convert_to_pandas,
    convert_to_polars,
    normalize_table,
)
from ngio.utils import NgioValueError, ZarrGroupHandler


class AbstractBaseTable(ABC):
    """Abstract base class for a table.

    This is used to define common methods and properties
    for all tables.

    This class is not meant to be used directly.
    """

    def __init__(
        self,
        table_data: TabularData | None = None,
        *,
        meta: BackendMeta | None = None,
    ) -> None:
        """Initialize the table."""
        if meta is None:
            meta = BackendMeta()

        self._meta = meta
        if table_data is not None:
            table_data = normalize_table(
                table_data,
                index_key=meta.index_key,
                index_type=meta.index_type,
            )
        self._table_data = table_data
        self._table_backend = None

    def __repr__(self) -> str:
        """Return a string representation of the table."""
        return f"{self.__class__.__name__}"

    @staticmethod
    @abstractmethod
    def table_type() -> str:
        """Return the type of the table."""
        ...

    @staticmethod
    @abstractmethod
    def version() -> str:
        """The generic table does not have a version.

        Since does not follow a specific schema.
        """
        ...

    @property
    def backend_name(self) -> str | None:
        """Return the name of the backend."""
        if self._table_backend is None:
            return None
        return self._table_backend.backend_name()

    @property
    def meta(self) -> BackendMeta:
        """Return the metadata of the table."""
        return self._meta

    @property
    def index_key(self) -> str | None:
        """Get the index key."""
        return self._meta.index_key

    @property
    def index_type(self) -> Literal["int", "str"] | None:
        """Get the index type."""
        return self._meta.index_type

    def load_as_anndata(self) -> AnnData:
        """Load the table as an AnnData object."""
        if self._table_backend is None:
            raise NgioValueError("No backend set for the table.")
        return self._table_backend.load_as_anndata()

    def load_as_pandas_df(self) -> pd.DataFrame:
        """Load the table as a pandas DataFrame."""
        if self._table_backend is None:
            raise NgioValueError("No backend set for the table.")
        return self._table_backend.load_as_pandas_df()

    def load_as_polars_lf(self) -> pl.LazyFrame:
        """Load the table as a polars LazyFrame."""
        if self._table_backend is None:
            raise NgioValueError("No backend set for the table.")
        return self._table_backend.load_as_polars_lf()

    @property
    def table_data(self) -> TabularData:
        """Return the table."""
        if self._table_data is not None:
            return self._table_data

        if self._table_backend is None:
            raise NgioValueError(
                "The table does not have a DataFrame in memory nor a backend."
            )

        self._table_data = self._table_backend.load()
        return self._table_data

    @property
    def dataframe(self) -> pd.DataFrame:
        """Return the table as a DataFrame."""
        return convert_to_pandas(
            self.table_data, index_key=self.index_key, index_type=self.index_type
        )

    @property
    def lazy_frame(self) -> pl.LazyFrame:
        """Return the table as a LazyFrame."""
        return convert_to_polars(
            self.table_data, index_key=self.index_key, index_type=self.index_type
        )

    @property
    def anndata(self) -> AnnData:
        """Return the table as an AnnData object."""
        return convert_to_anndata(self.table_data, index_key=self.index_key)

    @staticmethod
    def _load_backend(
        meta: BackendMeta,
        handler: ZarrGroupHandler,
        backend: TableBackend,
    ) -> TableBackendProtocol:
        """Create a new ROI table from a Zarr group handler."""
        if isinstance(backend, str):
            return ImplementedTableBackends().get_backend(
                backend_name=backend,
                group_handler=handler,
                index_key=meta.index_key,
                index_type=meta.index_type,
            )
        backend.set_group_handler(
            group_handler=handler,
            index_key=meta.index_key,
            index_type=meta.index_type,
        )
        return backend

    def set_table_data(
        self,
        table_data: TabularData | None = None,
        refresh: bool = False,
    ) -> None:
        """Set the table.

        If an object is passed, it will be used as the table.
        If None is passed, the table will be loaded from the backend.

        If refresh is True, the table will be reloaded from the backend.
            If table is not None, this will be ignored.
        """
        if table_data is not None:
            if not isinstance(table_data, TabularData):
                raise NgioValueError(
                    "The table must be a pandas DataFrame, polars LazyFrame, "
                    " or AnnData object."
                )

            self._table_data = normalize_table(
                table_data,
                index_key=self.index_key,
                index_type=self.index_type,
            )
            return None

        if self._table_data is not None and not refresh:
            return None

        if self._table_backend is None:
            raise NgioValueError(
                "The table does not have a DataFrame in memory nor a backend."
            )
        self._table_data = self._table_backend.load()

    def set_backend(
        self,
        handler: ZarrGroupHandler | None = None,
        backend: TableBackend = DefaultTableBackend,
    ) -> None:
        """Set the backend of the table."""
        if handler is None:
            if self._table_backend is None:
                raise NgioValueError(
                    "No backend set for the table yet. "
                    "A ZarrGroupHandler must be provided."
                )
            handler = self._table_backend.group_handler

        meta = self._meta
        _backend = self._load_backend(
            meta=meta,
            handler=handler,
            backend=backend,
        )
        self._table_backend = _backend

    @classmethod
    def _from_handler(
        cls,
        handler: ZarrGroupHandler,
        meta_model: builtins.type[BackendMeta],
        backend: TableBackend | None = None,
    ) -> Self:
        """Create a new ROI table from a Zarr group handler."""
        meta = meta_model(**handler.load_attrs())
        table = cls(meta=meta)
        if backend is None:
            backend = meta.backend
        table.set_backend(handler=handler, backend=backend)
        return table

    @classmethod
    @abstractmethod
    def from_handler(
        cls,
        handler: ZarrGroupHandler,
        backend: TableBackend | None = None,
    ) -> Self:
        """Create a new ROI table from a Zarr group handler."""
        pass

    @classmethod
    def from_table_data(cls, table_data: TabularData, meta: BackendMeta) -> Self:
        """Create a new ROI table from a Zarr group handler."""
        return cls(
            table_data=table_data,
            meta=meta,
        )

    def consolidate(self) -> None:
        """Write the current state of the table to the Zarr file."""
        if self._table_backend is None:
            raise NgioValueError(
                "No backend set for the table. "
                "Please add the table to a OME-Zarr Image before calling consolidate."
            )

        self._table_backend.write(
            self.table_data,
            metadata=self._meta.model_dump(exclude_none=True),
        )



from ngio.common._zoom import (
    InterpolationOrder,
)
from ngio.images._abstract_image import AbstractImage
from ngio.io_pipes._zoom_transform import BaseZoomTransform


class ZoomTransform(BaseZoomTransform):
    def __init__(
        self,
        input_image: AbstractImage,
        target_image: AbstractImage,
        order: InterpolationOrder = "nearest",
    ) -> None:
        super().__init__(
            input_dimensions=input_image.dimensions,
            target_dimensions=target_image.dimensions,
            order=order,
        )



from collections.abc import Sequence

import zarr

from ngio.common._dimensions import Dimensions
from ngio.common._roi import Roi
from ngio.io_pipes._io_pipes import (
    DaskGetter,
    DaskSetter,
    NumpyGetter,
    NumpySetter,
)
from ngio.io_pipes._ops_slices import SlicingInputType
from ngio.io_pipes._ops_transforms import TransformProtocol
from ngio.ome_zarr_meta.ngio_specs._pixel_size import PixelSize


def roi_to_slicing_dict(
    *,
    roi: Roi,
    pixel_size: PixelSize,
    slicing_dict: dict[str, SlicingInputType] | None = None,
) -> dict[str, SlicingInputType]:
    """Convert a ROI to a slicing dictionary."""
    roi_slicing_dict: dict[str, SlicingInputType] = roi.to_slicing_dict(
        pixel_size=pixel_size
    )  # type: ignore
    if slicing_dict is None:
        return roi_slicing_dict

    # Additional slice kwargs can be provided
    # and will override the ones from the ROI
    roi_slicing_dict.update(slicing_dict)
    return roi_slicing_dict


class NumpyRoiGetter(NumpyGetter):
    def __init__(
        self,
        *,
        zarr_array: zarr.Array,
        dimensions: Dimensions,
        roi: Roi,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        slicing_dict: dict[str, SlicingInputType] | None = None,
        remove_channel_selection: bool = False,
    ) -> None:
        input_slice_kwargs = roi_to_slicing_dict(
            roi=roi,
            pixel_size=dimensions.pixel_size,
            slicing_dict=slicing_dict,
        )
        super().__init__(
            zarr_array=zarr_array,
            dimensions=dimensions,
            axes_order=axes_order,
            transforms=transforms,
            slicing_dict=input_slice_kwargs,
            remove_channel_selection=remove_channel_selection,
            roi=roi,
        )


class DaskRoiGetter(DaskGetter):
    def __init__(
        self,
        *,
        zarr_array: zarr.Array,
        dimensions: Dimensions,
        roi: Roi,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        slicing_dict: dict[str, SlicingInputType] | None = None,
        remove_channel_selection: bool = False,
    ) -> None:
        input_slice_kwargs = roi_to_slicing_dict(
            roi=roi,
            pixel_size=dimensions.pixel_size,
            slicing_dict=slicing_dict,
        )
        super().__init__(
            zarr_array=zarr_array,
            dimensions=dimensions,
            axes_order=axes_order,
            transforms=transforms,
            slicing_dict=input_slice_kwargs,
            remove_channel_selection=remove_channel_selection,
            roi=roi,
        )


class NumpyRoiSetter(NumpySetter):
    def __init__(
        self,
        *,
        zarr_array: zarr.Array,
        dimensions: Dimensions,
        roi: Roi,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        slicing_dict: dict[str, SlicingInputType] | None = None,
        remove_channel_selection: bool = False,
    ) -> None:
        input_slice_kwargs = roi_to_slicing_dict(
            roi=roi,
            pixel_size=dimensions.pixel_size,
            slicing_dict=slicing_dict,
        )
        super().__init__(
            zarr_array=zarr_array,
            dimensions=dimensions,
            axes_order=axes_order,
            transforms=transforms,
            slicing_dict=input_slice_kwargs,
            remove_channel_selection=remove_channel_selection,
            roi=roi,
        )


class DaskRoiSetter(DaskSetter):
    def __init__(
        self,
        *,
        zarr_array: zarr.Array,
        dimensions: Dimensions,
        roi: Roi,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        slicing_dict: dict[str, SlicingInputType] | None = None,
        remove_channel_selection: bool = False,
    ) -> None:
        input_slice_kwargs = roi_to_slicing_dict(
            roi=roi,
            pixel_size=dimensions.pixel_size,
            slicing_dict=slicing_dict,
        )
        super().__init__(
            zarr_array=zarr_array,
            dimensions=dimensions,
            axes_order=axes_order,
            transforms=transforms,
            slicing_dict=input_slice_kwargs,
            remove_channel_selection=remove_channel_selection,
            roi=roi,
        )



# tests/stores/test_s3_store.py

from pathlib import Path

from utils import (
    check_ome_zarr,
    create_sample_ome_zarr,
    derive_image,
    get_s3_mapper,
    random_zarr_path,
)

S3_STORE_SUPPORTED_BACKENDS = ["anndata", "json"]
# CSV and Parquet work locally on S3 store, but not when using moto server for testing.
# To be investigated.


def test_s3_store(moto_s3_server: dict) -> None:
    # create boto3 client pointing at moto server
    endpoint_url, bucket_name = (
        moto_s3_server["endpoint_url"],
        moto_s3_server["bucket_name"],
    )
    store = get_s3_mapper(
        endpoint_url, bucket_name=bucket_name, zarr_path=random_zarr_path()
    )
    ome_zarr = create_sample_ome_zarr(
        store=store, supported_backends=S3_STORE_SUPPORTED_BACKENDS
    )
    check_ome_zarr(ome_zarr, supported_backends=S3_STORE_SUPPORTED_BACKENDS)


def test_s3_store_derive_to_s3_store(moto_s3_server: dict) -> None:
    endpoint_url, bucket_name = (
        moto_s3_server["endpoint_url"],
        moto_s3_server["bucket_name"],
    )
    store = get_s3_mapper(
        endpoint_url,
        bucket_name=bucket_name,
        zarr_path=random_zarr_path(),
    )
    ome_zarr = create_sample_ome_zarr(
        store=store, supported_backends=S3_STORE_SUPPORTED_BACKENDS
    )
    other_store = get_s3_mapper(
        endpoint_url,
        bucket_name=bucket_name,
        zarr_path=random_zarr_path(),
    )
    derived_ome_zarr = derive_image(ome_zarr, other_store=other_store)
    check_ome_zarr(derived_ome_zarr, supported_backends=S3_STORE_SUPPORTED_BACKENDS)


def test_s3_store_derive_to_local_store(moto_s3_server: dict, tmp_path: Path) -> None:
    endpoint_url, bucket_name = (
        moto_s3_server["endpoint_url"],
        moto_s3_server["bucket_name"],
    )
    store = get_s3_mapper(
        endpoint_url,
        bucket_name=bucket_name,
        zarr_path=random_zarr_path(),
    )
    ome_zarr = create_sample_ome_zarr(
        store=store, supported_backends=S3_STORE_SUPPORTED_BACKENDS
    )
    other_store = tmp_path / "s3_local_store_test" / random_zarr_path()
    derived_ome_zarr = derive_image(ome_zarr, other_store=other_store)
    check_ome_zarr(derived_ome_zarr, supported_backends=S3_STORE_SUPPORTED_BACKENDS)


def test_s3_store_derive_to_memory_store(moto_s3_server: dict) -> None:
    endpoint_url, bucket_name = (
        moto_s3_server["endpoint_url"],
        moto_s3_server["bucket_name"],
    )
    store = get_s3_mapper(
        endpoint_url,
        bucket_name=bucket_name,
        zarr_path=random_zarr_path(),
    )
    ome_zarr = create_sample_ome_zarr(
        store=store, supported_backends=S3_STORE_SUPPORTED_BACKENDS
    )
    other_store = {}
    derived_ome_zarr = derive_image(ome_zarr, other_store=other_store)
    check_ome_zarr(derived_ome_zarr, supported_backends=S3_STORE_SUPPORTED_BACKENDS)



from uuid import uuid4

import fsspec.implementations.http
import numpy as np
import pandas as pd
import s3fs

from ngio import OmeZarrContainer, create_empty_ome_zarr
from ngio.tables import FeatureTable

TEST_IMAGE = np.ones((3, 5, 64, 64), dtype=np.uint16)
TEST_LABEL = np.zeros((5, 64, 64), dtype=np.uint16)
TEST_LABEL[0, 10:30, 10:30] = 1
TEST_LABEL[0, 35:55, 35:55] = 2
TEST_LABEL[0, 20:40, 40:60] = 3

TEST_TABLE = pd.DataFrame(
    {
        "label": [1, 2, 3],
        "area": [150, 200, 250],
        "mean_intensity": [120.5, 130.0, 140.5],
    }
)
TEST_TABLE.set_index("label", inplace=True)


def set_to_image(ome_zarr: OmeZarrContainer) -> OmeZarrContainer:
    """Create and return an empty OME-Zarr container in the given store."""
    image = ome_zarr.get_image(path="1")
    image.set_array(patch=TEST_IMAGE)
    image.consolidate()
    return ome_zarr


def check_image_data(
    ome_zarr: OmeZarrContainer,
) -> None:
    """Retrieve the array data from the first image at the specified pyramid level."""
    image = ome_zarr.get_image()
    array_data = image.get_as_numpy()
    np.testing.assert_array_equal(array_data, TEST_IMAGE)


def derive_label(
    ome_zarr: OmeZarrContainer,
) -> OmeZarrContainer:
    """Derive a label array from the first image in the OME-Zarr container."""
    label_array = ome_zarr.derive_label(name="labels")
    label_array.set_array(patch=TEST_LABEL)
    label_array.consolidate()
    return ome_zarr


def check_label_data(
    ome_zarr: OmeZarrContainer,
) -> None:
    """Retrieve the label data from the specified label array."""
    label_array = ome_zarr.get_label(name="labels")
    label_data = label_array.get_as_numpy()
    np.testing.assert_array_equal(label_data, TEST_LABEL)


def add_table_to_ome_zarr(
    ome_zarr: OmeZarrContainer,
    backend: str = "anndata",
) -> OmeZarrContainer:
    """Add a table to the OME-Zarr container."""
    test_feature_table = FeatureTable(table_data=TEST_TABLE, reference_label="labels")
    name = f"test_table_{backend}"
    ome_zarr.add_table(
        name=name,
        table=test_feature_table,
        backend=backend,
    )
    return ome_zarr


def check_table_data(
    ome_zarr: OmeZarrContainer,
    table_name: str,
) -> None:
    """Retrieve the table data from the specified table in the OME-Zarr container."""
    table = ome_zarr.get_table(name=table_name)
    dataframe = table.dataframe
    pd.testing.assert_frame_equal(
        dataframe.sort_index(axis=1), TEST_TABLE.sort_index(axis=1)
    )


def derive_image(
    ome_zarr: OmeZarrContainer,
    other_store,
    copy_labels: bool = True,
    copy_tables: bool = True,
) -> OmeZarrContainer:
    """Derive a new image from the first image in the OME-Zarr container."""
    derived_ome_zarr = ome_zarr.derive_image(
        store=other_store,
        copy_labels=copy_labels,
        copy_tables=copy_tables,
    )
    return derived_ome_zarr


def create_sample_ome_zarr(
    store, supported_backends: list[str] | None = None
) -> OmeZarrContainer:
    """Create a sample OME-Zarr structure in the given store for testing."""
    ome_zarr = create_empty_ome_zarr(
        store=store,
        shape=(3, 5, 64, 64),
        pixelsize=(0.65, 0.65),
        channels_meta=["Channel 1", "Channel 2", "Channel 3"],
        levels=3,
        axes_names=["c", "z", "y", "x"],
    )
    ome_zarr = set_to_image(ome_zarr)
    check_image_data(ome_zarr)
    ome_zarr = derive_label(ome_zarr)
    if supported_backends is None:
        supported_backends = ["anndata", "json", "csv", "parquet"]

    for backend in supported_backends:
        ome_zarr = add_table_to_ome_zarr(ome_zarr, backend=backend)
    return ome_zarr


def check_ome_zarr(
    ome_zarr: OmeZarrContainer,
    check_tables: bool = True,
    check_labels: bool = True,
    supported_backends: list[str] | None = None,
) -> None:
    """Check that all tables in the OME-Zarr container match the test table."""
    if check_labels:
        check_label_data(ome_zarr)
    if check_tables:
        if supported_backends is None:
            supported_backends = ["anndata", "json", "csv", "parquet"]

        table_list = ome_zarr.list_tables()
        assert len(table_list) >= len(supported_backends)
        for table_name in table_list:
            backend = table_name.split("_")[-1]
            if backend in supported_backends:
                check_table_data(ome_zarr, table_name=table_name)


def get_s3_mapper(base_url: str, bucket_name: str, zarr_path: str):
    s3_fs = s3fs.S3FileSystem(
        key="test",
        secret="test",
        client_kwargs={
            "endpoint_url": base_url,
            "region_name": "us-east-1",
        },
    )
    return s3_fs.get_mapper(f"{bucket_name}/{zarr_path}")


def get_http_mapper(base_url: str, zarr_path: str):
    http_fs = fsspec.implementations.http.HTTPFileSystem()
    return http_fs.get_mapper(f"{base_url}/{zarr_path}")


def random_zarr_path() -> str:
    """Generate a random zarr path for testing."""
    return f"test_zarr_{uuid4()}.zarr"



from pathlib import Path
from typing import Literal

import numpy as np
import pytest
from scipy import ndimage
from skimage.segmentation import watershed

from ngio import create_ome_zarr_from_array, open_ome_zarr_container
from ngio.transforms import ZoomTransform


def _draw_random_labels(shape: tuple[int, ...], num_regions: int):
    np.random.seed(0)
    markers = np.zeros(shape, dtype=np.int32)
    seeds_list = np.random.randint(0, shape[0], size=(num_regions, 2))
    for i, (y, x) in enumerate(seeds_list, start=1):
        markers[y, x] = i

    image = ndimage.distance_transform_edt(markers == 0).astype("uint32")
    labels = watershed(image, markers).astype("uint32")
    return image, labels


@pytest.mark.parametrize(
    "shape",
    [(64, 64), (16, 32, 32)],
)
def test_get_masking(tmp_path: Path, shape: tuple[int, ...]):
    store = tmp_path / "test_image_yx_random_label.zarr"
    # Create a new ome_zarr with the mask
    mask, label_image = _draw_random_labels(shape=shape, num_regions=20)
    ome_zarr = create_ome_zarr_from_array(
        store=store,
        array=mask,
        xy_pixelsize=0.5,
        levels=2,
        overwrite=True,
    )
    full_res_image = ome_zarr.get_image(path="0")
    for level_path in ome_zarr.levels_paths:
        label_name = f"label_{level_path}"
        masking_table_name = f"label_ROI_table_{level_path}"
        image = ome_zarr.get_image(path=level_path)
        label = ome_zarr.derive_label(label_name, ref_image=image)
        zoom = ZoomTransform(
            input_image=label,
            target_image=full_res_image,
        )
        label.set_array(label_image, transforms=[zoom])
        label.consolidate()

        masking_roi = label.build_masking_roi_table()
        ome_zarr.add_table(masking_table_name, masking_roi)
        # Masking image test
        _ = ome_zarr.get_masked_image(masking_label_name=label_name)
        _ = ome_zarr.get_masked_image(masking_table_name=masking_table_name)
        _ = ome_zarr.get_masked_image(
            masking_table_name=masking_table_name, masking_label_name=label_name
        )
        _ = ome_zarr.get_masked_image(masking_label_name=label_name, path="1")


@pytest.mark.parametrize(
    "array_mode, shape",
    [("numpy", (64, 64)), ("dask", (64, 64)), ("numpy", (16, 32, 32))],
)
def test_masking(
    tmp_path: Path, array_mode: Literal["numpy", "dask"], shape: tuple[int, ...]
):
    mask, label_image = _draw_random_labels(shape=shape, num_regions=20)
    unique_labels, counts = np.unique(label_image, return_counts=True)
    labels_stats = dict(zip(unique_labels, counts, strict=True))

    store = tmp_path / "test_image_yx_random_label.zarr"
    # Create a new ome_zarr with the mask
    ome_zarr = create_ome_zarr_from_array(
        store=store,
        array=mask,
        xy_pixelsize=0.5,
        levels=1,
        overwrite=True,
    )
    label = ome_zarr.derive_label("label")
    label.set_array(label_image)

    # Masking image test
    masked_image = ome_zarr.get_masked_image("label")
    assert isinstance(masked_image.__repr__(), str)
    _roi_array = masked_image.get_roi(label=1, zoom_factor=1.123, mode=array_mode)
    masked_image.set_roi_masked(
        label=1, patch=np.ones_like(_roi_array), zoom_factor=1.123
    )

    _ = masked_image.get_roi_as_numpy(label=1)
    _ = masked_image.get_roi_as_dask(label=1)
    _ = masked_image.get_roi_masked_as_numpy(label=1)
    _ = masked_image.get_roi_masked_as_dask(label=1)

    _roi_mask = masked_image.get_roi_masked(label=1, mode=array_mode)
    # Check that the mask is binary after masking
    np.testing.assert_allclose(np.unique(_roi_mask), [0, 1])

    # Just test the API
    masked_image.set_roi(label=1, patch=np.zeros_like(_roi_array), zoom_factor=1.123)

    # Masking label test (recreate the label)
    ome_zarr.derive_label("empty_label")
    masked_new_label = ome_zarr.get_masked_label(
        "empty_label", masking_label_name="label"
    )
    assert isinstance(masked_new_label.__repr__(), str)

    for label_id in labels_stats.keys():
        label_mask = masked_new_label.get_roi(label_id, mode=array_mode)
        label_mask = np.full(label_mask.shape, label_id, dtype=label_mask.dtype)
        # Set the label only inside the mask
        masked_new_label.set_roi_masked(label_id, label_mask)

    # rerun the stats on the new masked label
    unique_labels, counts = np.unique(masked_new_label.get_array(), return_counts=True)
    labels_stats_masked = dict(zip(unique_labels, counts, strict=True))
    assert labels_stats == labels_stats_masked

    for label_id in labels_stats.keys():
        x = masked_new_label.get_roi_masked(label_id, mode=array_mode, zoom_factor=1.1)
        masked_new_label.set_roi(label_id, x, zoom_factor=1.1)


@pytest.mark.filterwarnings("ignore::anndata._warnings.ImplicitModificationWarning")
@pytest.mark.parametrize(
    ("label", "c", "zoom_factor", "expected_shape"),
    [
        (1009, (0,), 2, (1, 1, 154, 167)),
        (1009, 0, 2, (1, 154, 167)),
        (1009, 0, 1.123, (1, 86, 95)),
    ],
)
def test_real_mask(
    cardiomyocyte_small_mip_path: Path, label, c, zoom_factor, expected_shape
):
    # Test on a real example
    path = cardiomyocyte_small_mip_path / "B" / "03" / "0"
    ome_zarr = open_ome_zarr_container(path)
    masked_image = ome_zarr.get_masked_image("nuclei")
    image_data = masked_image.get_roi_masked_as_numpy(
        label=label, c=c, zoom_factor=zoom_factor
    )
    assert image_data.shape == expected_shape
    masked_image.set_roi_masked(
        label=label, patch=image_data, c=c, zoom_factor=zoom_factor
    )

    data_no_mask = masked_image.get_roi_as_numpy(
        label=label, c=c, zoom_factor=zoom_factor
    )
    assert data_no_mask.shape == image_data.shape



"""Download testing OME-Zarr datasets."""

import shutil
from pathlib import Path
from typing import Literal

import pooch

from ngio.utils._errors import NgioValueError


class UnzipAndRename(pooch.Unzip):
    """Unzip and rename the extracted directory."""

    def __init__(
        self,
        extract_dir: str = "",
        out_name: str = "ome-zarr.zarr",
        re_unzip: bool = True,
        **kwargs,
    ):
        super().__init__(extract_dir=extract_dir, **kwargs)
        self.out_name = out_name
        self.re_unzip = re_unzip

    def output_file(self) -> Path:
        """Return the output file path."""
        if self.extract_dir is None:
            raise NgioValueError("extract_dir must be set for UnzipAndRename.")

        return Path(self.extract_dir) / self.out_name

    def _extract_file(self, fname, extract_dir):
        """Extract the file and rename it."""
        output_path = self.output_file()
        if output_path.exists() and not self.re_unzip:
            # Nothing to do, the file already exists and we are not re-unzipping
            return None

        tmp_dir = Path(extract_dir) / "tmp"
        # If tmp_dir exists, remove it
        if tmp_dir.exists():
            shutil.rmtree(tmp_dir, ignore_errors=True)
        tmp_dir.mkdir(parents=True, exist_ok=True)

        super()._extract_file(fname, tmp_dir)

        list_extracted_dirs = tmp_dir.iterdir()
        # Keep only if ends with .zarr
        list_extracted_dirs = filter(
            lambda x: x.name.endswith(".zarr"),
            list_extracted_dirs,
        )
        list_extracted_dirs = list(list_extracted_dirs)
        if len(list_extracted_dirs) != 1:
            raise NgioValueError(
                "Expected one directory to be extracted, "
                f"got {len(list_extracted_dirs)}."
            )

        extracted_dir = list_extracted_dirs[0]
        if output_path.exists():
            shutil.rmtree(output_path, ignore_errors=True)

        extracted_dir.rename(output_path)
        # Clean up the temporary directory
        shutil.rmtree(tmp_dir, ignore_errors=True)


_ome_zarr_zoo = {
    "CardiomyocyteTiny": {
        "url": "https://zenodo.org/records/13305156/files/20200812-CardiomyocyteDifferentiation14-Cycle1.zarr.zip",
        "known_hash": "md5:efc21fe8d4ea3abab76226d8c166452c",
        "fname": "20200812-CardiomyocyteDifferentiation14-Cycle1-tiny.zarr.zip",
        "description": "Tiny cardiomyocyte dataset 3D (32MB).",
    },
    "CardiomyocyteTinyMip": {
        "url": "https://zenodo.org/records/13305156/files/20200812-CardiomyocyteDifferentiation14-Cycle1_mip.zarr.zip",
        "known_hash": "md5:51809479777cafbe9ac0f9fa5636aa95",
        "fname": "20200812-CardiomyocyteDifferentiation14-Cycle1-tiny-mip.zarr.zip",
        "description": "Tiny cardiomyocyte dataset 2D MIP (16.4MB).",
    },
    "CardiomyocyteSmall": {
        "url": "https://zenodo.org/records/13305316/files/20200812-CardiomyocyteDifferentiation14-Cycle1.zarr.zip",
        "known_hash": "md5:d5752ed4b72a9092a0290b3c04c0b9c2",
        "fname": "20200812-CardiomyocyteDifferentiation14-Cycle1-small.zarr.zip",
        "description": "Small cardiomyocyte dataset 3D (750MB).",
    },
    "CardiomyocyteSmallMip": {
        "url": "https://zenodo.org/records/13305316/files/20200812-CardiomyocyteDifferentiation14-Cycle1_mip.zarr.zip",
        "known_hash": "md5:3ed3ea898e0ed42d397da2e1dbe40750",
        "fname": "20200812-CardiomyocyteDifferentiation14-Cycle1-small-mip.zarr.zip",
        "description": "Small cardiomyocyte dataset 2D MIP (106MB).",
    },
    "CardiomyocyteMediumMip": {
        "url": "https://zenodo.org/records/14826000/files/20200812-CardiomyocyteDifferentiation14-Cycle1_mip.zarr.zip",
        "known_hash": "md5:3f932bbf7fc0577f58b97471707816a1",
        "fname": "20200812-CardiomyocyteDifferentiation14-Cycle1_mip.zarr.zip",
        "description": "Medium cardiomyocyte dataset 2D MIP (30GB).",
    },
}

AVAILABLE_DATASETS = Literal[
    "CardiomyocyteTiny",
    "CardiomyocyteTinyMip",
    "CardiomyocyteSmall",
    "CardiomyocyteSmallMip",
    "CardiomyocyteMediumMip",
]


def list_ome_zarr_datasets() -> list[str]:
    """List available OME-Zarr datasets."""
    return list(_ome_zarr_zoo.keys())


def print_datasets_infos() -> None:
    for dataset_name, dataset_info in _ome_zarr_zoo.items():
        print(f"{dataset_name} - Description: {dataset_info['description']}")


def download_ome_zarr_dataset(
    dataset_name: AVAILABLE_DATASETS | str,
    download_dir: str | Path = "data",
    re_unzip: bool = True,
    progressbar: bool = False,
) -> Path:
    """Download an OME-Zarr dataset.

    To list available datasets, use `list_ome_zarr_datasets`.

    Args:
        dataset_name (str): The dataset name.
        download_dir (str): The download directory. Defaults to "data".
        re_unzip (bool): If True, it will unzip the dataset even if it already exists.
        progressbar (bool): If True, show a progress bar during download.
    """
    if dataset_name not in _ome_zarr_zoo:
        raise NgioValueError(f"Dataset {dataset_name} not found in the OME-Zarr zoo.")
    zenodo_infos = _ome_zarr_zoo[dataset_name]

    fname = zenodo_infos["fname"]
    zarrname = fname.replace(".zip", "")

    processor = UnzipAndRename(
        extract_dir="",
        out_name=zarrname,
        re_unzip=re_unzip,
    )

    pooch.retrieve(
        url=zenodo_infos["url"],
        known_hash=zenodo_infos["known_hash"],
        fname=fname,
        path=download_dir,
        processor=processor,
        progressbar=progressbar,
    )
    return processor.output_file()



from collections.abc import Sequence
from typing import TypeVar

import dask.array as da
import numpy as np
from pydantic import BaseModel, ConfigDict

from ngio.common._dimensions import Dimensions
from ngio.utils import NgioValueError

##############################################################
#
# "AxesOps" Model
#
##############################################################


class AxesOps(BaseModel):
    """Model to represent axes operations.

    This model will be used to transform objects from on disk axes to in memory axes.
    """

    input_axes: tuple[str, ...]
    output_axes: tuple[str, ...]
    transpose_op: tuple[int, ...] | None = None
    expand_op: tuple[int, ...] | None = None
    squeeze_op: tuple[int, ...] | None = None
    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True)

    @property
    def is_no_op(self) -> bool:
        """Check if all operations are no ops."""
        if (
            self.transpose_op is None
            and self.expand_op is None
            and self.squeeze_op is None
        ):
            return True
        return False

    @property
    def get_transpose_op(self) -> tuple[int, ...] | None:
        """Get the transpose axes."""
        return self.transpose_op

    @property
    def get_expand_op(self) -> tuple[int, ...] | None:
        """Get the expand axes."""
        return self.expand_op

    @property
    def get_squeeze_op(self) -> tuple[int, ...] | None:
        """Get the squeeze axes."""
        return self.squeeze_op

    @property
    def set_transpose_op(self) -> tuple[int, ...] | None:
        """Set the transpose axes."""
        if self.transpose_op is None:
            return None
        return tuple(np.argsort(self.transpose_op))

    @property
    def set_expand_op(self) -> tuple[int, ...] | None:
        """Set the expand axes."""
        return self.squeeze_op

    @property
    def set_squeeze_op(self) -> tuple[int, ...] | None:
        """Set the squeeze axes."""
        return self.expand_op


##############################################################
#
# Axes Operations implementations
#
##############################################################


def _apply_numpy_axes_ops(
    array: np.ndarray,
    squeeze_axes: tuple[int, ...] | None = None,
    transpose_axes: tuple[int, ...] | None = None,
    expand_axes: tuple[int, ...] | None = None,
) -> np.ndarray:
    """Apply axes operations to a numpy array."""
    if squeeze_axes is not None:
        array = np.squeeze(array, axis=squeeze_axes)
    if transpose_axes is not None:
        array = np.transpose(array, axes=transpose_axes)
    if expand_axes is not None:
        array = np.expand_dims(array, axis=expand_axes)
    return array


def _apply_dask_axes_ops(
    array: da.Array,
    squeeze_axes: tuple[int, ...] | None = None,
    transpose_axes: tuple[int, ...] | None = None,
    expand_axes: tuple[int, ...] | None = None,
) -> da.Array:
    """Apply axes operations to a dask array."""
    if squeeze_axes is not None:
        array = da.squeeze(array, axis=squeeze_axes)
    if transpose_axes is not None:
        array = da.transpose(array, axes=transpose_axes)
    if expand_axes is not None:
        array = da.expand_dims(array, axis=expand_axes)
    return array


T = TypeVar("T")


def _apply_sequence_axes_ops(
    input_: Sequence[T],
    default: T,
    squeeze_axes: tuple[int, ...] | None = None,
    transpose_axes: tuple[int, ...] | None = None,
    expand_axes: tuple[int, ...] | None = None,
) -> list[T]:
    input_list = list(input_)
    if squeeze_axes is not None:
        for offset, ax in enumerate(squeeze_axes):
            input_list.pop(ax - offset)

    if transpose_axes is not None:
        input_list = [input_list[i] for i in transpose_axes]

    if expand_axes is not None:
        for ax in expand_axes:
            input_list.insert(ax, default)

    return input_list


def get_as_numpy_axes_ops(
    array: np.ndarray,
    axes_ops: AxesOps,
) -> np.ndarray:
    """Apply axes operations to a numpy array."""
    return _apply_numpy_axes_ops(
        array,
        squeeze_axes=axes_ops.get_squeeze_op,
        transpose_axes=axes_ops.get_transpose_op,
        expand_axes=axes_ops.get_expand_op,
    )


def get_as_dask_axes_ops(
    array: da.Array,
    axes_ops: AxesOps,
) -> da.Array:
    """Apply axes operations to a dask array."""
    return _apply_dask_axes_ops(
        array,
        squeeze_axes=axes_ops.get_squeeze_op,
        transpose_axes=axes_ops.get_transpose_op,
        expand_axes=axes_ops.get_expand_op,
    )


def get_as_sequence_axes_ops(
    input_: Sequence[T],
    axes_ops: AxesOps,
    default: T,
) -> list[T]:
    """Apply axes operations to a sequence."""
    return _apply_sequence_axes_ops(
        input_,
        default=default,
        squeeze_axes=axes_ops.get_squeeze_op,
        transpose_axes=axes_ops.get_transpose_op,
        expand_axes=axes_ops.get_expand_op,
    )


def set_as_numpy_axes_ops(
    array: np.ndarray,
    axes_ops: AxesOps,
) -> np.ndarray:
    """Apply inverse axes operations to a numpy array."""
    return _apply_numpy_axes_ops(
        array,
        squeeze_axes=axes_ops.set_squeeze_op,
        transpose_axes=axes_ops.set_transpose_op,
        expand_axes=axes_ops.set_expand_op,
    )


def set_as_dask_axes_ops(
    array: da.Array,
    axes_ops: AxesOps,
) -> da.Array:
    """Apply inverse axes operations to a dask array."""
    return _apply_dask_axes_ops(
        array,
        squeeze_axes=axes_ops.set_squeeze_op,
        transpose_axes=axes_ops.set_transpose_op,
        expand_axes=axes_ops.set_expand_op,
    )


def set_as_sequence_axes_ops(
    input_: Sequence[T],
    axes_ops: AxesOps,
    default: T,
) -> list[T]:
    """Apply inverse axes operations to a sequence."""
    return _apply_sequence_axes_ops(
        input_,
        default=default,
        squeeze_axes=axes_ops.set_squeeze_op,
        transpose_axes=axes_ops.set_transpose_op,
        expand_axes=axes_ops.set_expand_op,
    )


##############################################################
#
# Builder functions
#
##############################################################


def _check_output_axes(axes: Sequence[str]) -> None:
    """Check that the input axes are valid."""
    unique_names = set(axes)
    if len(unique_names) != len(axes):
        raise NgioValueError(
            "Duplicate axis names found. Please provide unique names for each axis."
        )
    for name in axes:
        if not isinstance(name, str):
            raise NgioValueError(
                f"Invalid axis name '{name}'. Axis names must be strings."
            )


def _build_squeeze_tuple(
    input_axes: tuple[str, ...], output_axes: tuple[str, ...]
) -> tuple[tuple[int, ...], tuple[str, ...]]:
    """Build a tuple of axes to squeeze."""
    axes_to_squeeze = []
    axes_after_squeeze = []
    for i, ax in enumerate(input_axes):
        if ax not in output_axes:
            axes_to_squeeze.append(i)
        else:
            axes_after_squeeze.append(ax)
    return tuple(axes_to_squeeze), tuple(axes_after_squeeze)


def _build_transpose_tuple(
    input_axes: tuple[str, ...], output_axes: tuple[str, ...]
) -> tuple[tuple[int, ...], tuple[str, ...]]:
    """Build a tuple of axes to transpose."""
    transposition_order = []
    axes_names_after_transpose = []
    for ax in output_axes:
        if ax in input_axes:
            transposition_order.append(input_axes.index(ax))
            axes_names_after_transpose.append(ax)
    return tuple(transposition_order), tuple(axes_names_after_transpose)


def _build_expand_tuple(
    input_axes: tuple[str, ...], output_axes: tuple[str, ...]
) -> tuple[int, ...]:
    """Build a tuple of axes to expand."""
    axes_to_expand = []
    for i, ax in enumerate(output_axes):
        if ax not in input_axes:
            axes_to_expand.append(i)
    return tuple(axes_to_expand)


def _build_axes_ops(
    input_axes: tuple[str, ...], output_axes: tuple[str, ...]
) -> AxesOps:
    """Change the order of the axes."""
    # Validate the names
    _check_output_axes(output_axes)
    # Step 1: Check find squeeze axes
    axes_to_squeeze, input_axes = _build_squeeze_tuple(input_axes, output_axes)
    # Step 2: Find the transposition order
    transposition_order, input_axes = _build_transpose_tuple(input_axes, output_axes)
    # Step 3: Find axes to expand
    axes_to_expand = _build_expand_tuple(input_axes, output_axes)

    # If the operations are empty, make them None
    if len(axes_to_squeeze) == 0:
        axes_to_squeeze = None

    if np.allclose(transposition_order, np.arange(len(transposition_order))):
        # If the transposition order is the identity, we don't need to transpose
        transposition_order = None
    if len(axes_to_expand) == 0:
        axes_to_expand = None

    return AxesOps(
        input_axes=input_axes,
        output_axes=output_axes,
        transpose_op=transposition_order,
        expand_op=axes_to_expand,
        squeeze_op=axes_to_squeeze,
    )


def _normalize_axes_order(
    dimensions: Dimensions,
    axes_order: Sequence[str],
) -> tuple[str, ...]:
    """Convert axes order to the on-disk axes names.

    In this way there is not unambiguity in the axes order.
    """
    new_axes_order = []
    for axis_name in axes_order:
        axis = dimensions.axes_handler.get_axis(axis_name)
        if axis is None:
            new_axes_order.append(axis_name)
        else:
            new_axes_order.append(axis.name)
    return tuple(new_axes_order)


def build_axes_ops(
    *,
    dimensions: Dimensions,
    input_axes: tuple[str, ...],
    axes_order: Sequence[str] | None,
) -> AxesOps:
    if axes_order is None:
        return AxesOps(
            input_axes=input_axes,
            output_axes=input_axes,
        )
    output_axes = _normalize_axes_order(dimensions=dimensions, axes_order=axes_order)

    axes_ops = _build_axes_ops(input_axes=input_axes, output_axes=output_axes)
    return axes_ops



from ngio.tables.backends._py_arrow_backends import PyArrowBackend


class CsvTableBackend(PyArrowBackend):
    """A class to load and write small tables in CSV format."""

    def __init__(
        self,
    ):
        """Initialize the CsvTableBackend."""
        super().__init__(
            table_name="table.csv",
            table_format="csv",
        )

    @staticmethod
    def backend_name() -> str:
        """Return the name of the backend."""
        return "csv"



"""Utility functions for converting between different tables formats.

The supported formats are:
- pandas DataFrame
- polars DataFrame or LazyFrame
- AnnData

These functions are used to validate and normalize the tables
to ensure that conversion between formats is consistent.
"""

from copy import deepcopy
from typing import Literal

import numpy as np
import pandas as pd
import pandas.api.types as ptypes
import polars as pl
from anndata import AnnData
from pandas import DataFrame
from polars import DataFrame as PolarsDataFrame
from polars import LazyFrame

from ngio.utils import NgioTableValidationError, NgioValueError

TabularData = AnnData | DataFrame | PolarsDataFrame | LazyFrame

# -----------------
# Validation utils
# -----------------


def _validate_index_key_df(pandas_df: DataFrame, index_key: str | None) -> DataFrame:
    """Validate the index key of the pandas DataFrame.

    Args:
        pandas_df (DataFrame): The pandas DataFrame to validate.
        index_key (str | None): The column name to use as the index of the DataFrame.

    Returns:
        DataFrame: DataFrame with validated index key.

    Raises:
        NgioTableValidationError: If index key is not found in DataFrame.
    """
    if index_key is None:
        return pandas_df

    if pandas_df.index.name == index_key:
        return pandas_df

    if index_key in pandas_df.columns:
        pandas_df = pandas_df.set_index(index_key)
        pandas_df.index.name = index_key
        return pandas_df

    if pandas_df.index.name is None:
        pandas_df.index.name = index_key
        return pandas_df

    raise NgioTableValidationError(f"Index key '{index_key}' is not found in DataFrame")


def _validate_cast_index_dtype_df(
    pandas_df: DataFrame, index_type: str | None
) -> DataFrame:
    """Check if the index of the DataFrame has the correct dtype.

    Args:
        pandas_df (DataFrame): The pandas DataFrame to validate.
        index_type (str | None): The type to cast the index to ('str' or 'int').

    Returns:
        DataFrame: DataFrame with index of the specified type.

    Raises:
        NgioTableValidationError: If index cannot be cast to the specified type.
        NgioValueError: If index_type is not 'str' or 'int'.
    """
    if index_type is None:
        # Nothing to do
        return pandas_df

    if index_type == "str":
        if ptypes.is_integer_dtype(pandas_df.index):
            # Convert the int index to string is generally safe
            pandas_df = pandas_df.set_index(pandas_df.index.astype(str))

        if not ptypes.is_string_dtype(pandas_df.index):
            raise NgioTableValidationError(
                f"Table index must be of string type, got {pandas_df.index.dtype}"
            )

    elif index_type == "int":
        if ptypes.is_string_dtype(pandas_df.index):
            # Try to convert the string index to int
            try:
                pandas_df = pandas_df.set_index(pandas_df.index.astype(int))
            except ValueError as e:
                if "invalid literal for int() with base 10" in str(e):
                    raise NgioTableValidationError(
                        "Table index must be of integer type, got str."
                        f" We tried implicit conversion and failed: {e}"
                    ) from None
                else:
                    raise e from e

        if not ptypes.is_integer_dtype(pandas_df.index):
            raise NgioTableValidationError(
                f"Table index must be of integer type, got {pandas_df.index.dtype}"
            )
    else:
        raise NgioValueError(
            f"Invalid index type '{index_type}'. Must be 'int' or 'str'."
        )

    return pandas_df


def _check_for_mixed_types(series: pd.Series) -> None:
    """Check if the column has mixed types.

    Args:
        series (pd.Series): The pandas Series to check.

    Raises:
        NgioTableValidationError: If the column has mixed types.
    """
    if series.apply(type).nunique() > 1:  # type: ignore (type lint fails here)
        raise NgioTableValidationError(
            f"Column {series.name} has mixed types: "
            f"{series.apply(type).unique()}. "  # type: ignore (type lint fails here)
            "Type of all elements must be the same."
        )


def _check_for_supported_types(series: pd.Series) -> Literal["str", "int", "numeric"]:
    """Check if the column has supported types.

    Args:
        series (pd.Series): The pandas Series to check.

    Returns:
        Literal["str", "int", "numeric"]: The type category of the series.

    Raises:
        NgioTableValidationError: If the column has unsupported types.
    """
    if ptypes.is_string_dtype(series):
        return "str"
    if ptypes.is_integer_dtype(series):
        return "int"
    if ptypes.is_numeric_dtype(series):
        return "numeric"
    raise NgioTableValidationError(
        f"Column {series.name} has unsupported type: {series.dtype}."
        " Supported types are string and numerics."
    )


# -----------------
# Normalization functions
# -----------------


def normalize_pandas_df(
    pandas_df: DataFrame,
    index_key: str | None = None,
    index_type: Literal["int", "str"] | None = None,
    reset_index: bool = False,
) -> DataFrame:
    """Make sure the DataFrame has the correct index and dtype.

    Args:
        pandas_df (DataFrame): The pandas DataFrame to validate.
        index_key (str | None): The column name to use as the index of the DataFrame.
            Default is None.
        index_type (str | None): The type of the index column in the DataFrame.
            Either 'str' or 'int'. Default is None.
        reset_index (bool): If True the index will be reset (i.e. the index will be
            converted to a column). If False, the index will be kept as is.

    Returns:
        DataFrame: Normalized pandas DataFrame.
    """
    pandas_df = _validate_index_key_df(pandas_df, index_key)
    pandas_df = _validate_cast_index_dtype_df(pandas_df, index_type)
    if pandas_df.index.name is not None:
        index_key = str(pandas_df.index.name)

    if reset_index and pandas_df.index.name is not None:
        pandas_df = pandas_df.reset_index()
    return pandas_df


def normalize_polars_lf(
    polars_lf: LazyFrame | PolarsDataFrame,
    index_key: str | None = None,
    index_type: Literal["int", "str"] | None = None,
) -> LazyFrame:
    """Validate the polars LazyFrame.

    Args:
        polars_lf (LazyFrame | PolarsDataFrame): The polars LazyFrame to validate.
        index_key (str | None): The column name to use as the index of the DataFrame.
            Default is None.
        index_type (str | None): The type of the index column in the DataFrame.
            Either 'str' or 'int'. Default is None.

    Returns:
        LazyFrame: Normalized polars LazyFrame.

    Raises:
        ValueError: If index_key is not found or index_type is invalid.
    """
    if index_key is not None:
        schema = polars_lf.collect_schema()
        if index_key not in schema:
            raise NgioTableValidationError(
                f"Index key '{index_key}' not found in LazyFrame columns."
            )

        if index_type is not None:
            if index_type not in ["int", "str"]:
                raise NgioTableValidationError(
                    f"Invalid index type '{index_type}'. Must be 'int' or 'str'."
                )
            if index_type == "int" and not schema[index_key].is_integer():
                polars_lf = polars_lf.with_columns(pl.col(index_key).cast(pl.Int64))
            elif index_type == "str" and not schema[index_key] == pl.String():
                polars_lf = polars_lf.with_columns(pl.col(index_key).cast(pl.String()))

    if isinstance(polars_lf, PolarsDataFrame):
        polars_lf = polars_lf.lazy()
    return polars_lf


def normalize_anndata(
    anndata: AnnData,
    index_key: str | None = None,
) -> AnnData:
    """Validate the AnnData object.

    Args:
        anndata (AnnData): The AnnData object to validate.
        index_key (str | None): The column name to use as the index of the DataFrame.
            Default is None.

    Returns:
        AnnData: Normalized AnnData object.
    """
    if index_key is None:
        return anndata
    obs = _validate_index_key_df(anndata.obs, index_key)
    obs = _validate_cast_index_dtype_df(obs, "str")

    if obs.equals(anndata.obs):
        return anndata

    anndata = deepcopy(anndata)
    anndata.obs = obs
    return anndata


# -----------------
# Conversion functions
# -----------------


def convert_pandas_to_polars(
    pandas_df: DataFrame,
    index_key: str | None = None,
    index_type: Literal["int", "str"] | None = None,
) -> LazyFrame:
    """Convert a pandas DataFrame to a polars LazyFrame.

    Args:
        pandas_df (DataFrame): The pandas DataFrame to convert.
        index_key (str | None): The column name to use as the index of the DataFrame.
            Default is None.
        index_type (str | None): The type of the index column in the DataFrame.
            Either 'str' or 'int'. Default is None.

    Returns:
        LazyFrame: Converted and normalized polars LazyFrame.
    """
    pandas_df = normalize_pandas_df(
        pandas_df,
        index_key=index_key,
        index_type=index_type,
        reset_index=True,
    )
    return pl.from_pandas(pandas_df).lazy()


def convert_polars_to_pandas(
    polars_df: PolarsDataFrame | LazyFrame,
    index_key: str | None = None,
    index_type: Literal["int", "str"] | None = None,
    reset_index: bool = False,
) -> DataFrame:
    """Convert a polars DataFrame or LazyFrame to a pandas DataFrame.

    Args:
        polars_df (PolarsDataFrame | LazyFrame): The polars DataFrame or
            LazyFrame to convert.
        index_key (str | None): The column name to use as the index of the DataFrame.
            Default is None.
        index_type (str | None): The type of the index column in the DataFrame.
            Either 'str' or 'int'. Default is None.
        reset_index (bool): If True the index will be reset (i.e., the index will be
            converted to a column). If False, the index will be kept as is.

    Returns:
        DataFrame: Converted and normalized pandas DataFrame.
    """
    if isinstance(polars_df, LazyFrame):
        polars_df = polars_df.collect()

    pandas_df = polars_df.to_pandas()
    pandas_df = normalize_pandas_df(
        pandas_df,
        index_key=index_key,
        index_type=index_type,
        reset_index=reset_index,
    )
    return pandas_df


def convert_pandas_to_anndata(
    pandas_df: DataFrame,
    index_key: str | None = None,
) -> AnnData:
    """Convert a pandas DataFrame to an AnnData object.

    Args:
        pandas_df (DataFrame): The pandas DataFrame to convert.
        index_key (str | None): The column name to use as the index of the DataFrame.
            Default is None.

    Returns:
        AnnData: Converted AnnData object.
    """
    pandas_df = normalize_pandas_df(
        pandas_df,
        index_key=index_key,
        index_type="str",
        reset_index=False,
    )

    str_columns, int_columns, num_columns = [], [], []
    for col_name in pandas_df.columns:
        column = pandas_df[col_name]
        _check_for_mixed_types(column)  # Mixed types are not allowed in the table
        col_type = _check_for_supported_types(
            column
        )  # Only string and numeric types are allowed

        if col_type == "str":
            str_columns.append(col_name)

        elif col_type == "int":
            int_columns.append(col_name)

        elif col_type == "numeric":
            num_columns.append(col_name)

    # Converting all observations to string
    obs_df = pandas_df[str_columns + int_columns]
    obs_df.index = pandas_df.index

    x_df = pandas_df[num_columns]

    if x_df.dtypes.nunique() > 1:
        x_df = x_df.astype("float64")

    if x_df.empty:
        # If there are no numeric columns, create an empty array
        # to avoid AnnData failing to create the object
        x_df = np.zeros((len(obs_df), 0), dtype="float64")

    return AnnData(X=x_df, obs=obs_df)


def convert_anndata_to_pandas(
    anndata: AnnData,
    index_key: str | None = None,
    index_type: Literal["int", "str"] | None = None,
    reset_index: bool = False,
) -> DataFrame:
    """Convert an AnnData object to a pandas DataFrame.

    Args:
        anndata (AnnData): An AnnData object to convert.
        index_key (str | None): The column name to use as the index of the DataFrame.
            Default is None.
        index_type (str | None): The type of the index column in the DataFrame.
            Either 'str' or 'int'. Default is None.
        reset_index (bool): If True the index will be reset (i.e., the index will be
            converted to a column). If False, the index will be kept as is.

    Returns:
        DataFrame: Converted and normalized pandas DataFrame.
    """
    pandas_df = anndata.to_df()
    pandas_df[anndata.obs.columns.to_list()] = anndata.obs
    pandas_df = normalize_pandas_df(
        pandas_df,
        index_key=index_key,
        index_type=index_type,
        reset_index=reset_index,
    )
    return pandas_df


def convert_anndata_to_polars(
    anndata: AnnData,
    index_key: str | None = None,
    index_type: Literal["int", "str"] | None = None,
) -> LazyFrame:
    """Convert an AnnData object to a polars LazyFrame.

    Args:
        anndata (AnnData): An AnnData object to convert.
        index_key (str | None): The column name to use as the index of the DataFrame.
            Default is None.
        index_type (str | None): The type of the index column in the DataFrame.
            Either 'str' or 'int'. Default is None.

    Returns:
        LazyFrame: Converted and normalized polars LazyFrame.
    """
    pandas_df = convert_anndata_to_pandas(
        anndata,
        index_key=index_key,
        index_type=index_type,
        reset_index=True,
    )
    return pl.from_pandas(pandas_df).lazy()


def convert_polars_to_anndata(
    polars_df: LazyFrame | PolarsDataFrame,
    index_key: str | None = None,
) -> AnnData:
    """Convert a polars LazyFrame or DataFrame to an AnnData object.

    Args:
        polars_df (LazyFrame | PolarsDataFrame): The polars LazyFrame or
            DataFrame to convert.
        index_key (str | None): The column name to use as the index of the DataFrame.
            Default is None.

    Returns:
        AnnData: Converted AnnData object.
    """
    if isinstance(polars_df, LazyFrame):
        polars_df = polars_df.collect()
    pandas_df = polars_df.to_pandas()
    return convert_pandas_to_anndata(
        pandas_df,
        index_key=index_key,
    )


# -----------------
# Conversion functions
# -----------------


def normalize_table(
    table_data: TabularData,
    index_key: str | None = None,
    index_type: Literal["int", "str"] | None = None,
) -> TabularData:
    """Normalize a table to a specific format.

    Args:
        table_data (TabularData): The table to normalize.
        index_key (str | None): The column name to use as the index of the DataFrame.
            Default is None.
        index_type (str | None): The type of the index column in the DataFrame.
            Either 'str' or 'int'. Default is None.

    Returns:
        DataFrame | AnnData | PolarsDataFrame | LazyFrame: Normalized table.
    """
    if isinstance(table_data, DataFrame):
        return normalize_pandas_df(
            table_data,
            index_key=index_key,
            index_type=index_type,
            reset_index=False,
        )
    if isinstance(table_data, AnnData):
        return normalize_anndata(table_data, index_key=index_key)
    if isinstance(table_data, PolarsDataFrame) or isinstance(table_data, LazyFrame):
        return normalize_polars_lf(
            table_data,
            index_key=index_key,
            index_type=index_type,
        )
    raise NgioValueError(f"Unsupported table type: {type(table_data)}")


def convert_to_anndata(
    table_data: TabularData,
    index_key: str | None = None,
) -> AnnData:
    """Convert a table to an AnnData object.

    Args:
        table_data (TabularData): The table to convert.
        index_key (str | None): The column name to use as the index of the DataFrame.
            Default is None.

    Returns:
        AnnData: Converted AnnData object.
    """
    if isinstance(table_data, AnnData):
        return normalize_anndata(table_data, index_key=index_key)
    if isinstance(table_data, DataFrame):
        return convert_pandas_to_anndata(table_data, index_key=index_key)
    if isinstance(table_data, PolarsDataFrame) or isinstance(table_data, LazyFrame):
        return convert_polars_to_anndata(table_data, index_key=index_key)
    raise NgioValueError(f"Unsupported table type: {type(table_data)}")


def convert_to_pandas(
    table_data: TabularData,
    index_key: str | None = None,
    index_type: Literal["int", "str"] | None = None,
    reset_index: bool = False,
) -> DataFrame:
    """Convert a table to a pandas DataFrame.

    Args:
        table_data (TabularData): The table to convert.
        index_key (str | None): The column name to use as the index of the DataFrame.
            Default is None.
        index_type (str | None): The type of the index column in the DataFrame.
            Either 'str' or 'int'. Default is None.
        reset_index (bool): If True the index will be reset (i.e., the index will be
            converted to a column). If False, the index will be kept as is.

    Returns:
        DataFrame: Converted pandas DataFrame.
    """
    if isinstance(table_data, DataFrame):
        return normalize_pandas_df(
            table_data,
            index_key=index_key,
            index_type=index_type,
            reset_index=reset_index,
        )
    if isinstance(table_data, AnnData):
        return convert_anndata_to_pandas(
            table_data,
            index_key=index_key,
            index_type=index_type,
            reset_index=reset_index,
        )
    if isinstance(table_data, PolarsDataFrame) or isinstance(table_data, LazyFrame):
        return convert_polars_to_pandas(
            table_data,
            index_key=index_key,
            index_type=index_type,
            reset_index=reset_index,
        )
    raise NgioValueError(f"Unsupported table type: {type(table_data)}")


def convert_to_polars(
    table_data: TabularData,
    index_key: str | None = None,
    index_type: Literal["int", "str"] | None = None,
) -> LazyFrame:
    """Convert a table to a polars LazyFrame.

    Args:
        table_data (TabularData): The table to convert.
        index_key (str | None): The column name to use as the index of the DataFrame.
            Default is None.
        index_type (str | None): The type of the index column in the DataFrame.
            Either 'str' or 'int'. Default is None.

    Returns:
        LazyFrame: Converted polars LazyFrame.
    """
    if isinstance(table_data, PolarsDataFrame) or isinstance(table_data, LazyFrame):
        return normalize_polars_lf(
            table_data,
            index_key=index_key,
            index_type=index_type,
        )
    if isinstance(table_data, DataFrame):
        return convert_pandas_to_polars(
            table_data,
            index_key=index_key,
            index_type=index_type,
        )
    if isinstance(table_data, AnnData):
        return convert_anndata_to_polars(
            table_data,
            index_key=index_key,
            index_type=index_type,
        )
    raise NgioValueError(f"Unsupported table type: {type(table_data)}")



from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from anndata import AnnData

from ngio.tables._tables_container import open_table, write_table
from ngio.tables.v1 import GenericTable


@pytest.mark.parametrize("backend", ["json", "anndata"])
def test_generic_df_table(tmp_path: Path, backend: str):
    store = tmp_path / "test.zarr"
    test_df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    table = GenericTable(table_data=test_df)
    assert isinstance(table.__repr__(), str)

    write_table(store=store, table=table, backend=backend)

    loaded_table = open_table(store=store)
    assert isinstance(loaded_table, GenericTable)
    assert table.backend_name == backend

    assert set(loaded_table.dataframe.columns) == {"a", "b"}
    for column in loaded_table.dataframe.columns:
        pd.testing.assert_series_equal(
            loaded_table.dataframe[column], test_df[column], check_index=False
        )

    loaded_table.load_as_pandas_df()
    loaded_table.load_as_polars_lf()
    loaded_table.load_as_anndata()


@pytest.mark.parametrize("backend", ["anndata"])
def test_generic_anndata_table(tmp_path: Path, backend: str):
    store = tmp_path / "test.zarr"
    test_df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    test_obs = pd.DataFrame({"c": ["a", "b", "c"]})
    test_obs.index = test_obs.index.astype(str)
    test_obsm = np.random.normal(0, 1, size=(3, 2))

    anndata = AnnData(X=test_df, obs=test_obs)
    anndata.obsm["test"] = test_obsm

    table = GenericTable(table_data=anndata)

    assert isinstance(table.table_data, AnnData)

    write_table(store=store, table=table, backend=backend)

    loaded_table = open_table(store=store)
    assert isinstance(loaded_table, GenericTable)

    loaded_ad = loaded_table.load_as_anndata()
    loaded_df = loaded_table.dataframe
    assert set(loaded_df.columns) == {"a", "b", "c"}

    np.testing.assert_allclose(loaded_ad.obsm["test"], test_obsm)  # type: ignore



from pathlib import Path

from utils import (
    check_ome_zarr,
    create_sample_ome_zarr,
    derive_image,
    get_s3_mapper,
    random_zarr_path,
)

MEMORY_STORE_SUPPORTED_BACKENDS = ["anndata", "json", "csv", "parquet"]


def test_memory_store() -> None:
    store_path = {}
    ome_zarr = create_sample_ome_zarr(
        store=store_path, supported_backends=MEMORY_STORE_SUPPORTED_BACKENDS
    )
    check_ome_zarr(ome_zarr, supported_backends=MEMORY_STORE_SUPPORTED_BACKENDS)


def test_memory_store_derive_to_memory_store() -> None:
    store_path = {}
    ome_zarr = create_sample_ome_zarr(
        store=store_path, supported_backends=MEMORY_STORE_SUPPORTED_BACKENDS
    )
    other_store = {}
    derived_ome_zarr = derive_image(ome_zarr, other_store=other_store)
    check_ome_zarr(derived_ome_zarr, supported_backends=["anndata", "json"])


def test_memory_store_derive_to_local_store(tmp_path: Path) -> None:
    store_path = {}
    ome_zarr = create_sample_ome_zarr(
        store=store_path, supported_backends=MEMORY_STORE_SUPPORTED_BACKENDS
    )
    other_store = tmp_path / "local_store_test" / "derived_test.zarr"
    derived_ome_zarr = derive_image(ome_zarr, other_store=other_store)
    check_ome_zarr(derived_ome_zarr, supported_backends=["anndata", "json"])


def test_memory_store_derive_to_s3_store(moto_s3_server: dict) -> None:
    store_path = {}
    ome_zarr = create_sample_ome_zarr(
        store=store_path, supported_backends=MEMORY_STORE_SUPPORTED_BACKENDS
    )
    # create boto3 client pointing at moto server
    other_store = get_s3_mapper(
        base_url=moto_s3_server["endpoint_url"],
        bucket_name=moto_s3_server["bucket_name"],
        zarr_path=random_zarr_path(),
    )
    derived_ome_zarr = derive_image(ome_zarr, other_store=other_store)
    check_ome_zarr(derived_ome_zarr, supported_backends=["anndata", "json"])



"""Next Generation file format IO."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("ngio")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "uninstalled"
__author__ = "Lorenzo Cerrone"
__email__ = "lorenzo.cerrone@uzh.ch"

from ngio.common import Dimensions, Roi, RoiSlice
from ngio.hcs import (
    OmeZarrPlate,
    OmeZarrWell,
    create_empty_plate,
    create_empty_well,
    open_ome_zarr_plate,
    open_ome_zarr_well,
)
from ngio.images import (
    ChannelSelectionModel,
    Image,
    Label,
    OmeZarrContainer,
    create_empty_ome_zarr,
    create_ome_zarr_from_array,
    create_synthetic_ome_zarr,
    open_image,
    open_label,
    open_ome_zarr_container,
)
from ngio.ome_zarr_meta.ngio_specs import (
    AxesSetup,
    DefaultNgffVersion,
    ImageInWellPath,
    NgffVersions,
    PixelSize,
)

__all__ = [
    "AxesSetup",
    "ChannelSelectionModel",
    "DefaultNgffVersion",
    "Dimensions",
    "Image",
    "ImageInWellPath",
    "Label",
    "NgffVersions",
    "OmeZarrContainer",
    "OmeZarrPlate",
    "OmeZarrWell",
    "PixelSize",
    "Roi",
    "RoiSlice",
    "create_empty_ome_zarr",
    "create_empty_plate",
    "create_empty_well",
    "create_ome_zarr_from_array",
    "create_synthetic_ome_zarr",
    "open_image",
    "open_label",
    "open_ome_zarr_container",
    "open_ome_zarr_plate",
    "open_ome_zarr_well",
]



"""Region of interest (ROI) metadata.

These are the interfaces bwteen the ROI tables / masking ROI tables and
    the ImageLikeHandler.
"""

from collections.abc import Callable
from typing import Literal, Self

from pydantic import BaseModel, ConfigDict, Field, field_validator

from ngio.ome_zarr_meta import PixelSize
from ngio.utils import NgioValueError


def world_to_pixel(value: float, pixel_size: float, eps: float = 1e-6) -> float:
    raster_value = value / pixel_size

    # If the value is very close to an integer, round it
    # This ensures that we don't have floating point precision issues
    # When loading ROIs that were originally defined in pixel coordinates
    _rounded = round(raster_value)
    if abs(_rounded - raster_value) < eps:
        return _rounded
    return raster_value


def pixel_to_world(value: float, pixel_size: float) -> float:
    return value * pixel_size


def _join_roi_names(name1: str | None, name2: str | None) -> str | None:
    if name1 is not None and name2 is not None:
        if name1 == name2:
            return name1
        return f"{name1}:{name2}"
    return name1 or name2


def _join_roi_labels(label1: int | None, label2: int | None) -> int | None:
    if label1 is not None and label2 is not None:
        if label1 == label2:
            return label1
        raise NgioValueError("Cannot join ROIs with different labels")
    return label1 or label2


class RoiSlice(BaseModel):
    axis_name: str
    start: float | None = Field(default=None)
    length: float | None = Field(default=None, ge=0)

    model_config = ConfigDict(extra="forbid")

    @classmethod
    def _from_slice(
        cls,
        axis_name: str,
        selection: slice,
    ) -> "RoiSlice":
        start = selection.start
        length = (
            None
            if selection.stop is None or selection.start is None
            else selection.stop - selection.start
        )
        return cls(axis_name=axis_name, start=start, length=length)

    @classmethod
    def from_value(
        cls,
        axis_name: str,
        value: float | tuple[float | None, float | None] | slice,
    ) -> "RoiSlice":
        if isinstance(value, slice):
            return cls._from_slice(axis_name=axis_name, selection=value)
        elif isinstance(value, tuple):
            return cls(axis_name=axis_name, start=value[0], length=value[1])
        elif isinstance(value, int | float):
            return cls(axis_name=axis_name, start=value, length=1)
        else:
            raise TypeError(f"Unsupported type for slice value: {type(value)}")

    def __repr__(self) -> str:
        return f"{self.axis_name}: {self.start}->{self.end}"

    @property
    def end(self) -> float | None:
        if self.start is None or self.length is None:
            return None
        return self.start + self.length

    def to_slice(self) -> slice:
        return slice(self.start, self.end)

    def _is_compatible(self, other: "RoiSlice", msg: str) -> None:
        if self.axis_name != other.axis_name:
            raise NgioValueError(
                f"{msg}: Cannot operate on RoiSlices with different axis names"
            )

    def union(self, other: "RoiSlice") -> "RoiSlice":
        self._is_compatible(other, "RoiSlice union failed")
        start = min(self.start or 0, other.start or 0)
        end = max(self.end or float("inf"), other.end or float("inf"))
        length = end - start if end > start else 0
        if length == float("inf"):
            length = None
        return RoiSlice(axis_name=self.axis_name, start=start, length=length)

    def intersection(self, other: "RoiSlice") -> "RoiSlice | None":
        self._is_compatible(other, "RoiSlice intersection failed")
        start = max(self.start or 0, other.start or 0)
        end = min(self.end or float("inf"), other.end or float("inf"))
        if end <= start:
            # No intersection
            return None
        length = end - start
        if length == float("inf"):
            length = None
        return RoiSlice(axis_name=self.axis_name, start=start, length=length)

    def to_world(self, pixel_size: float) -> "RoiSlice":
        start = (
            pixel_to_world(self.start, pixel_size) if self.start is not None else None
        )
        length = (
            pixel_to_world(self.length, pixel_size) if self.length is not None else None
        )
        return RoiSlice(axis_name=self.axis_name, start=start, length=length)

    def to_pixel(self, pixel_size: float) -> "RoiSlice":
        start = (
            world_to_pixel(self.start, pixel_size) if self.start is not None else None
        )
        length = (
            world_to_pixel(self.length, pixel_size) if self.length is not None else None
        )
        return RoiSlice(axis_name=self.axis_name, start=start, length=length)

    def zoom(self, zoom_factor: float = 1.0) -> "RoiSlice":
        if zoom_factor <= 0:
            raise NgioValueError("Zoom factor must be greater than 0")
        zoom_factor -= 1.0
        if self.length is None:
            return self

        diff_length = self.length * zoom_factor
        length = self.length + diff_length
        start = max((self.start or 0) - (diff_length / 2), 0)
        return RoiSlice(axis_name=self.axis_name, start=start, length=length)


class Roi(BaseModel):
    name: str | None
    slices: list[RoiSlice] = Field(min_length=2)
    label: int | None = Field(default=None, ge=0)
    space: Literal["world", "pixel"] = "world"

    model_config = ConfigDict(extra="allow")

    @field_validator("slices")
    @classmethod
    def validate_no_duplicate_axes(cls, v: list[RoiSlice]) -> list[RoiSlice]:
        axis_names = [s.axis_name for s in v]
        if len(axis_names) != len(set(axis_names)):
            raise NgioValueError("Roi slices must have unique axis names")
        return v

    def _nice_repr__(self) -> str:
        slices_repr = ", ".join(repr(s) for s in self.slices)
        if self.label is None:
            label_str = ""
        else:
            label_str = f", label={self.label}"

        if self.name is None:
            name_str = ""
        else:
            name_str = f"name={self.name}, "
        return f"Roi({name_str}{slices_repr}{label_str}, space={self.space})"

    @classmethod
    def from_values(
        cls,
        slices: dict[str, float | tuple[float | None, float | None] | slice],
        name: str | None,
        label: int | None = None,
        space: Literal["world", "pixel"] = "world",
        **kwargs,
    ) -> Self:
        _slices = []
        for axis, _slice in slices.items():
            _slices.append(RoiSlice.from_value(axis_name=axis, value=_slice))
        return cls.model_construct(
            name=name, slices=_slices, label=label, space=space, **kwargs
        )

    def get(self, axis_name: str) -> RoiSlice | None:
        for roi_slice in self.slices:
            if roi_slice.axis_name == axis_name:
                return roi_slice
        return None

    def get_name(self) -> str:
        if self.name is not None:
            return self.name
        if self.label is not None:
            return str(self.label)
        return self._nice_repr__()

    @staticmethod
    def _apply_sym_ops(
        self_slices: list[RoiSlice],
        other_slices: list[RoiSlice],
        op: Callable[[RoiSlice, RoiSlice], RoiSlice | None],
    ) -> list[RoiSlice] | None:
        self_axis_dict = {s.axis_name: s for s in self_slices}
        other_axis_dict = {s.axis_name: s for s in other_slices}
        common_axis_names = self_axis_dict.keys() | other_axis_dict.keys()
        new_slices = []
        for axis_name in common_axis_names:
            slice_a = self_axis_dict.get(axis_name)
            slice_b = other_axis_dict.get(axis_name)
            if slice_a is not None and slice_b is not None:
                result = op(slice_a, slice_b)
                if result is None:
                    return None
                new_slices.append(result)
            elif slice_a is not None:
                new_slices.append(slice_a)
            elif slice_b is not None:
                new_slices.append(slice_b)
        return new_slices

    def intersection(self, other: Self) -> Self | None:
        if self.space != other.space:
            raise NgioValueError(
                "Roi intersection failed: One ROI is in pixel space and the "
                "other in world space"
            )

        out_slices = self._apply_sym_ops(
            self.slices, other.slices, op=lambda a, b: a.intersection(b)
        )
        if out_slices is None:
            return None

        name = _join_roi_names(self.name, other.name)
        label = _join_roi_labels(self.label, other.label)
        return self.model_copy(
            update={"name": name, "slices": out_slices, "label": label}
        )

    def union(self, other: Self) -> Self:
        if self.space != other.space:
            raise NgioValueError(
                "Roi union failed: One ROI is in pixel space and the "
                "other in world space"
            )

        out_slices = self._apply_sym_ops(
            self.slices, other.slices, op=lambda a, b: a.union(b)
        )
        if out_slices is None:
            raise NgioValueError("Roi union failed: could not compute union")

        name = _join_roi_names(self.name, other.name)
        label = _join_roi_labels(self.label, other.label)
        return self.model_copy(
            update={"name": name, "slices": out_slices, "label": label}
        )

    def zoom(
        self, zoom_factor: float = 1.0, axes: tuple[str, ...] = ("x", "y")
    ) -> Self:
        new_slices = []
        for roi_slice in self.slices:
            if roi_slice.axis_name in axes:
                new_slices.append(roi_slice.zoom(zoom_factor=zoom_factor))
            else:
                new_slices.append(roi_slice)
        return self.model_copy(update={"slices": new_slices})

    def to_world(self, pixel_size: PixelSize | None = None) -> Self:
        if self.space == "world":
            return self.model_copy()
        if pixel_size is None:
            raise NgioValueError(
                "Pixel sizes must be provided to convert ROI from pixel to world"
            )
        new_slices = []
        for roi_slice in self.slices:
            pixel_size_ = pixel_size.get(roi_slice.axis_name, default=1.0)
            new_slices.append(roi_slice.to_world(pixel_size=pixel_size_))
        return self.model_copy(update={"slices": new_slices, "space": "world"})

    def to_pixel(self, pixel_size: PixelSize | None = None) -> Self:
        if self.space == "pixel":
            return self.model_copy()

        if pixel_size is None:
            raise NgioValueError(
                "Pixel sizes must be provided to convert ROI from world to pixel"
            )

        new_slices = []
        for roi_slice in self.slices:
            pixel_size_ = pixel_size.get(roi_slice.axis_name, default=1.0)
            new_slices.append(roi_slice.to_pixel(pixel_size=pixel_size_))
        return self.model_copy(update={"slices": new_slices, "space": "pixel"})

    def to_slicing_dict(self, pixel_size: PixelSize | None = None) -> dict[str, slice]:
        roi = self.to_pixel(pixel_size=pixel_size)
        return {roi_slice.axis_name: roi_slice.to_slice() for roi_slice in roi.slices}



"""Module for handling the /tables group in an OME-NGFF file."""

from typing import Literal, Protocol, TypeVar

import anndata as ad
import pandas as pd
import polars as pl

from ngio.tables.backends import (
    BackendMeta,
    DefaultTableBackend,
    TableBackend,
    TabularData,
)
from ngio.tables.v1 import (
    ConditionTableV1,
    FeatureTableV1,
    GenericTable,
    MaskingRoiTableV1,
    RoiTableV1,
)
from ngio.tables.v1._roi_table import GenericRoiTableV1
from ngio.utils import (
    AccessModeLiteral,
    NgioValidationError,
    NgioValueError,
    StoreOrGroup,
    ZarrGroupHandler,
)

GenericRoiTable = GenericRoiTableV1
RoiTable = RoiTableV1
MaskingRoiTable = MaskingRoiTableV1
FeatureTable = FeatureTableV1
ConditionTable = ConditionTableV1


class Table(Protocol):
    """Placeholder class for a table."""

    @staticmethod
    def table_type() -> str:
        """Return the type of the table."""
        ...

    @staticmethod
    def version() -> str:
        """Return the version of the table."""
        ...

    @property
    def backend_name(self) -> str | None:
        """The name of the backend."""
        ...

    @property
    def meta(self) -> BackendMeta:
        """Return the metadata for the table."""
        ...

    @property
    def dataframe(self) -> pd.DataFrame:
        """Return the table as a DataFrame."""
        ...

    @property
    def lazy_frame(self) -> pl.LazyFrame:
        """Return the table as a LazyFrame."""
        ...

    @property
    def anndata(self) -> ad.AnnData:
        """Return the table as an AnnData object."""
        ...

    def set_table_data(
        self,
        table_data: TabularData | None = None,
        refresh: bool = False,
    ) -> None:
        """Make sure that the table data is set (exist in memory).

        If an object is passed, it will be used as the table.
        If None is passed, the table will be loaded from the backend.

        If refresh is True, the table will be reloaded from the backend.
            If table is not None, this will be ignored.
        """
        ...

    def set_backend(
        self,
        handler: ZarrGroupHandler | None = None,
        backend: TableBackend = DefaultTableBackend,
    ) -> None:
        """Set the backend store and path for the table.

        Either a handler or a backend must be provided.

        If the handler in none it will be inferred from the backend.
        If the backend is none, it will be inferred from the group attrs
        """
        ...

    @classmethod
    def from_handler(
        cls,
        handler: ZarrGroupHandler,
        backend: TableBackend | None = None,
    ) -> "Table":
        """Create a new table from a Zarr group handler."""
        ...

    @classmethod
    def from_table_data(cls, table_data: TabularData, meta: BackendMeta) -> "Table":
        """Create a new table from a DataFrame."""
        ...

    @property
    def table_data(self) -> TabularData:
        """Return the table."""
        ...

    def consolidate(self) -> None:
        """Consolidate the table on disk."""
        ...


TypedTable = Literal[
    "generic_table",
    "roi_table",
    "masking_roi_table",
    "feature_table",
    "condition_table",
]

TypedRoiTable = Literal[
    "roi_table",
    "masking_roi_table",
]

TableType = TypeVar("TableType", bound=Table)


class TableMeta(BackendMeta):
    """Base class for table metadata."""

    table_version: str = "1"
    type: str = "generic_table"

    def unique_name(self) -> str:
        """Return the unique name for the table."""
        return f"{self.type}_v{self.table_version}"


def _get_meta(handler: ZarrGroupHandler) -> TableMeta:
    """Get the metadata from the handler."""
    attrs = handler.load_attrs()
    meta = TableMeta(**attrs)
    return meta


class ImplementedTables:
    """A singleton class to manage the available table handler plugins."""

    _instance = None
    _implemented_tables: dict[str, type[Table]]

    def __new__(cls):
        """Create a new instance of the class if it does not exist."""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._implemented_tables = {}
        return cls._instance

    def available_implementations(self) -> list[str]:
        """Get the available table handler versions."""
        return list(self._implemented_tables.keys())

    def get_table(
        self,
        meta: TableMeta,
        handler: ZarrGroupHandler,
        backend: TableBackend | None = None,
        strict: bool = True,
    ) -> Table:
        """Try to get a handler for the given store based on the metadata version."""
        if strict:
            default = None
        else:
            default = GenericTable

        table_cls = self._implemented_tables.get(meta.unique_name(), default)
        if table_cls is None:
            raise NgioValueError(
                f"Table handler for {meta.unique_name()} not implemented."
            )
        table = table_cls.from_handler(handler=handler, backend=backend)
        return table

    def _add_implementation(
        self, handler: type[Table], name: str, overwrite: bool = False
    ) -> None:
        """Register a new table handler."""
        if name in self._implemented_tables and not overwrite:
            raise NgioValueError(
                f"Table handler for {name} already implemented. "
                "Use overwrite=True to replace it."
            )
        self._implemented_tables[name] = handler

    def add_implementation(
        self,
        handler: type[Table],
        overwrite: bool = False,
        aliases: list[str] | None = None,
    ) -> None:
        """Register a new table handler."""
        meta = TableMeta(
            type=handler.table_type(),
            table_version=handler.version(),
        )

        self._add_implementation(handler, meta.unique_name(), overwrite)

        if aliases is not None:
            for alias in aliases:
                self._add_implementation(handler, alias, overwrite)


class TablesContainer:
    """A class to handle the /labels group in an OME-NGFF file."""

    def __init__(self, group_handler: ZarrGroupHandler) -> None:
        """Initialize the LabelGroupHandler."""
        self._group_handler = group_handler

        # Validate the group
        # Either contains a tables attribute or is empty
        attrs = self._group_handler.load_attrs()
        if len(attrs) == 0:
            # It's an empty group
            pass
        elif "tables" in attrs and isinstance(attrs["tables"], list):
            # It's a valid group
            pass
        else:
            raise NgioValidationError(
                f"Invalid /tables group. "
                f"Expected a single tables attribute with a list of table names. "
                f"Found: {attrs}"
            )

    def _get_tables_list(self) -> list[str]:
        """Create the /tables group if it doesn't exist."""
        attrs = self._group_handler.load_attrs()
        return attrs.get("tables", [])

    def _get_table_group_handler(self, name: str) -> ZarrGroupHandler:
        """Get the group handler for a table."""
        handler = self._group_handler.get_handler(path=name)
        return handler

    def list(self, filter_types: TypedTable | str | None = None) -> list[str]:
        """List all labels in the group."""
        tables = self._get_tables_list()
        if filter_types is None:
            return tables

        filtered_tables = []
        for table_name in tables:
            tb_handler = self._get_table_group_handler(table_name)
            table_type = _get_meta(tb_handler).type
            if table_type == filter_types:
                filtered_tables.append(table_name)
        return filtered_tables

    def get(
        self,
        name: str,
        backend: TableBackend | None = None,
        strict: bool = True,
    ) -> Table:
        """Get a label from the group."""
        if name not in self.list():
            raise NgioValueError(f"Table '{name}' not found in the group.")

        table_handler = self._get_table_group_handler(name)

        meta = _get_meta(table_handler)
        return ImplementedTables().get_table(
            meta=meta,
            handler=table_handler,
            backend=backend,
            strict=strict,
        )

    def get_as(
        self,
        name: str,
        table_cls: type[TableType],
        backend: TableBackend | None = None,
    ) -> TableType:
        """Get a table from the group as a specific type."""
        if name not in self.list():
            raise NgioValueError(f"Table '{name}' not found in the group.")

        table_handler = self._get_table_group_handler(name)
        return table_cls.from_handler(
            handler=table_handler,
            backend=backend,
        )  # type: ignore[return-value]

    def add(
        self,
        name: str,
        table: Table,
        backend: TableBackend = DefaultTableBackend,
        overwrite: bool = False,
    ) -> None:
        """Add a table to the group."""
        existing_tables = self._get_tables_list()
        if name in existing_tables and not overwrite:
            raise NgioValueError(
                f"Table '{name}' already exists in the group. "
                "Use overwrite=True to replace it."
            )

        table_handler = self._group_handler.get_handler(path=name, overwrite=overwrite)

        if backend is None:
            backend = table.backend_name

        table.set_table_data()
        table.set_backend(
            handler=table_handler,
            backend=backend,
        )
        table.consolidate()
        if name not in existing_tables:
            existing_tables.append(name)
            self._group_handler.write_attrs({"tables": existing_tables})


ImplementedTables().add_implementation(RoiTableV1)
ImplementedTables().add_implementation(MaskingRoiTableV1)
ImplementedTables().add_implementation(FeatureTableV1)
ImplementedTables().add_implementation(ConditionTableV1)

###################################################################################
#
# Utility functions to open and write tables
#
###################################################################################


def open_tables_container(
    store: StoreOrGroup,
    cache: bool = False,
    mode: AccessModeLiteral = "r+",
) -> TablesContainer:
    """Open a table handler from a Zarr store."""
    handler = ZarrGroupHandler(store=store, cache=cache, mode=mode)
    return TablesContainer(handler)


def open_table(
    store: StoreOrGroup,
    backend: TableBackend | None = None,
    cache: bool = False,
    mode: AccessModeLiteral = "r+",
) -> Table:
    """Open a table from a Zarr store."""
    handler = ZarrGroupHandler(
        store=store,
        cache=cache,
        mode=mode,
    )
    meta = _get_meta(handler)
    return ImplementedTables().get_table(
        meta=meta, handler=handler, backend=backend, strict=False
    )


def open_table_as(
    store: StoreOrGroup,
    table_cls: type[TableType],
    backend: TableBackend | None = None,
    cache: bool = False,
    mode: AccessModeLiteral = "r+",
) -> TableType:
    """Open a table from a Zarr store as a specific type."""
    handler = ZarrGroupHandler(
        store=store,
        cache=cache,
        mode=mode,
    )
    return table_cls.from_handler(
        handler=handler,
        backend=backend,
    )  # type: ignore[return-value]


def write_table(
    store: StoreOrGroup,
    table: Table,
    backend: TableBackend = DefaultTableBackend,
    cache: bool = False,
    mode: AccessModeLiteral = "a",
) -> None:
    """Write a table to a Zarr store.

    A table will be created at the given store location.

    Args:
        store (StoreOrGroup): The Zarr store or group to write the table to.
        table (Table): The table to write.
        backend (TableBackend): The backend to use for writing the table.
        cache (bool): Whether to use caching for the Zarr group handler.
        mode (AccessModeLiteral): The access mode to use for the Zarr group handler.

    """
    handler = ZarrGroupHandler(store=store, cache=cache, mode=mode)
    table.set_backend(
        handler=handler,
        backend=backend,
    )
    table.consolidate()



from pathlib import Path

import pandas as pd
import pytest

from ngio.tables._tables_container import open_table, write_table
from ngio.tables.v1 import FeatureTableV1


@pytest.mark.parametrize("backend", ["json", "anndata"])
def test_feature_table(tmp_path: Path, backend: str):
    store = tmp_path / "test.zarr"
    test_df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6], "label": [1, 2, 3]})
    table = FeatureTableV1(test_df, reference_label="label")
    assert isinstance(table.__repr__(), str)
    meta_dict = table.meta.model_dump()
    assert meta_dict.get("table_version") == table.version()
    assert meta_dict.get("type") == table.table_type()
    assert meta_dict.get("region") == {"path": "../labels/label"}
    assert table.reference_label == "label"

    write_table(store=store, table=table, backend=backend)

    loaded_table = open_table(store=store)
    assert isinstance(loaded_table, FeatureTableV1)
    assert set(loaded_table.dataframe.columns) == {"a", "b"}
    for column in loaded_table.dataframe.columns:
        pd.testing.assert_series_equal(
            loaded_table.dataframe[column], test_df[column], check_index=False
        )
    meta_dict = loaded_table.meta.model_dump()

    assert meta_dict.get("table_version") == loaded_table.version()
    assert meta_dict.get("type") == loaded_table.table_type()
    assert meta_dict.get("region") == {"path": "../labels/label"}
    assert table.reference_label == "label"



"""I/O pipes for reading and writing data from zarr to numpy and dask arrays.

There are 3 main types of I/O pipes:
- Standard I/O pipes: NumpyGetter, NumpySetter, DaskGetter, DaskSetter:
    These pipes read and write data from simple integer indexing and slicing.
- ROI I/O pipes: NumpyRoiGetter, NumpyRoiSetter, DaskRoiGetter, DaskRoiSetter:
    These pipes read and write data from a region of interest (ROI) defined in physical
    coordinates.
- Masked I/O pipes: NumpyGetterMasked, NumpySetterMasked, DaskGetterMasked,
    DaskSetterMasked: These pipes like the ROI pipes read and write data
    from a region of interest (ROI). However they also load a boolean mask
    from a label zarr array to mask the data being read or written.

All the io pipes are structured in the same way.

When reading data the order of operations is:
- Step 1: Slice the zarr array to load only the data needed into memory.
- Step 2: Apply axes operations to reorder, squeeze or expand the axes.
    To match the user desired axes order.
- Step 3: Apply any additional transforms to the data.

When writing data the order of operations is the reverse.

The Transforms must implement the TransformProtocol.
They should be stateless and only depend on the input array and the slicing
and axes ops. This allows them to be easily reused between different I/O pipes.

"""

from ngio.io_pipes._io_pipes import (
    DaskGetter,
    DaskSetter,
    DataGetter,
    DataSetter,
    NumpyGetter,
    NumpySetter,
)
from ngio.io_pipes._io_pipes_masked import (
    DaskGetterMasked,
    DaskSetterMasked,
    NumpyGetterMasked,
    NumpySetterMasked,
)
from ngio.io_pipes._io_pipes_roi import (
    DaskRoiGetter,
    DaskRoiSetter,
    NumpyRoiGetter,
    NumpyRoiSetter,
)
from ngio.io_pipes._match_shape import dask_match_shape, numpy_match_shape
from ngio.io_pipes._ops_slices import SlicingInputType, SlicingOps, SlicingType
from ngio.io_pipes._ops_transforms import TransformProtocol

__all__ = [
    "DaskGetter",
    "DaskGetterMasked",
    "DaskRoiGetter",
    "DaskRoiSetter",
    "DaskSetter",
    "DaskSetterMasked",
    "DataGetter",
    "DataSetter",
    "NumpyGetter",
    "NumpyGetterMasked",
    "NumpyRoiGetter",
    "NumpyRoiSetter",
    "NumpySetter",
    "NumpySetterMasked",
    "SlicingInputType",
    "SlicingOps",
    "SlicingType",
    "TransformProtocol",
    "dask_match_shape",
    "numpy_match_shape",
]



from typing import Literal

import polars as pl
import pyarrow as pa
import pyarrow.csv as pa_csv
import pyarrow.dataset as pa_ds
import pyarrow.fs as pa_fs
import pyarrow.parquet as pa_parquet
from pandas import DataFrame
from polars import DataFrame as PolarsDataFrame
from polars import LazyFrame
from zarr.storage import FsspecStore, LocalStore, MemoryStore, ZipStore

from ngio.tables.backends._abstract_backend import AbstractTableBackend
from ngio.tables.backends._utils import normalize_pandas_df, normalize_polars_lf
from ngio.utils import NgioValueError
from ngio.utils._zarr_utils import _make_sync_fs


class PyArrowBackend(AbstractTableBackend):
    """A class to load and write small tables in CSV format."""

    def __init__(
        self,
        table_name: str,
        table_format: Literal["csv", "parquet"] = "parquet",
    ):
        self.table_name = table_name
        self.table_format = table_format

    @staticmethod
    def implements_anndata() -> bool:
        """Whether the handler implements the anndata protocol."""
        return False

    @staticmethod
    def implements_pandas() -> bool:
        """Whether the handler implements the dataframe protocol."""
        return True

    @staticmethod
    def implements_polars() -> bool:
        """Whether the handler implements the polars protocol."""
        return True

    @staticmethod
    def backend_name() -> str:
        """Return the name of the backend."""
        raise NotImplementedError(
            "The backend_name method must be implemented in the subclass."
        )

    def _raise_store_type_not_supported(self):
        """Raise an error for unsupported store types."""
        ext = self.table_name.split(".")[-1]
        store = self._group_handler.store
        raise NgioValueError(
            f"Ngio does not support reading a {ext} table from a "
            f"store of type {type(store)}. "
            "Please make sure to use a compatible "
            "store like a LocalStore, or "
            "FsspecStore, or MemoryStore, or ZipStore."
        )

    def _load_from_local_store(self, store: LocalStore, path: str) -> pa_ds.Dataset:
        """Load the table from a directory store."""
        root_path = store.root
        table_path = f"{root_path}/{path}/{self.table_name}"
        dataset = pa_ds.dataset(table_path, format=self.table_format)
        return dataset

    def _load_from_fsspec_store(self, store: FsspecStore, path: str) -> pa_ds.Dataset:
        """Load the table from an FS store."""
        table_path = f"{store.path}/{path}/{self.table_name}"
        fs = _make_sync_fs(store.fs)
        dataset = pa_ds.dataset(table_path, format=self.table_format, filesystem=fs)
        return dataset

    def _load_from_in_memory_store(
        self, store: MemoryStore, path: str
    ) -> pa_ds.Dataset:
        """Load the table from an in-memory store."""
        table_path = f"{path}/{self.table_name}"
        table = store._store_dict.get(table_path, None)
        if table is None:
            raise NgioValueError(
                f"Table {self.table_name} not found in the in-memory store at "
                f"path {path}."
            )
        assert isinstance(table, pa.Table)
        dataset = pa_ds.dataset(table)
        return dataset

    def _load_from_zip_store(self, store: ZipStore, path: str) -> pa_ds.Dataset:
        """Load the table from a zip store."""
        raise NotImplementedError("Zip store loading is not implemented yet.")

    def _load_pyarrow_dataset(self) -> pa_ds.Dataset:
        """Load the table as a pyarrow Dataset."""
        store = self._group_handler.store
        path = self._group_handler.group.path
        if isinstance(store, LocalStore):
            return self._load_from_local_store(store, path)
        elif isinstance(store, FsspecStore):
            return self._load_from_fsspec_store(store, path)
        elif isinstance(store, MemoryStore):
            return self._load_from_in_memory_store(store, path)
        elif isinstance(store, ZipStore):
            return self._load_from_zip_store(store, path)
        self._raise_store_type_not_supported()

    def load_as_pandas_df(self) -> DataFrame:
        """Load the table as a pandas DataFrame."""
        dataset = self._load_pyarrow_dataset()
        dataframe = dataset.to_table().to_pandas()
        dataframe = normalize_pandas_df(
            dataframe,
            index_key=self.index_key,
            index_type=self.index_type,
            reset_index=False,
        )
        return dataframe

    def load(self) -> DataFrame:
        """Load the table as a pandas DataFrame."""
        return self.load_as_pandas_df()

    def load_as_polars_lf(self) -> LazyFrame:
        """Load the table as a polars LazyFrame."""
        dataset = self._load_pyarrow_dataset()
        lazy_frame = pl.scan_pyarrow_dataset(dataset)
        if not isinstance(lazy_frame, LazyFrame):
            raise NgioValueError(
                "Table is not a lazy frame. Please report this issue as an ngio bug."
                f" {type(lazy_frame)}"
            )

        lazy_frame = normalize_polars_lf(
            lazy_frame,
            index_key=self.index_key,
            index_type=self.index_type,
        )
        return lazy_frame

    def _write_to_stream(self, stream, table: pa.Table) -> None:
        """Write the table to a stream."""
        if self.table_format == "parquet":
            pa_parquet.write_table(table, stream)
        elif self.table_format == "csv":
            pa_csv.write_csv(table, stream)
        else:
            raise NgioValueError(
                f"Unsupported table format: {self.table_format}. "
                "Supported formats are 'parquet' and 'csv'."
            )

    def _write_to_local_store(
        self, store: LocalStore, path: str, table: pa.Table
    ) -> None:
        """Write the table to a directory store."""
        root_path = store.root
        table_path = f"{root_path}/{path}/{self.table_name}"
        self._write_to_stream(table_path, table)

    def _write_to_fsspec_store(
        self, store: FsspecStore, path: str, table: pa.Table
    ) -> None:
        """Write the table to an FS store."""
        table_path = f"{store.path}/{path}/{self.table_name}"
        fs = _make_sync_fs(store.fs)
        fs = pa_fs.PyFileSystem(pa_fs.FSSpecHandler(fs))
        with fs.open_output_stream(table_path) as out_stream:
            self._write_to_stream(out_stream, table)

    def _write_to_in_memory_store(
        self, store: MemoryStore, path: str, table: pa.Table
    ) -> None:
        """Write the table to an in-memory store."""
        table_path = f"{path}/{self.table_name}"
        store._store_dict[table_path] = table

    def _write_to_zip_store(self, store: ZipStore, path: str, table: pa.Table) -> None:
        """Write the table to a zip store."""
        raise NotImplementedError("Writing to zip store is not implemented yet.")

    def _write_pyarrow_dataset(self, dataset: pa.Table) -> None:
        """Write the table from a pyarrow Dataset."""
        store = self._group_handler.store
        path = self._group_handler.group.path
        if isinstance(store, LocalStore):
            return self._write_to_local_store(store=store, path=path, table=dataset)
        elif isinstance(store, FsspecStore):
            return self._write_to_fsspec_store(store=store, path=path, table=dataset)
        elif isinstance(store, MemoryStore):
            return self._write_to_in_memory_store(store=store, path=path, table=dataset)
        elif isinstance(store, ZipStore):
            return self._write_to_zip_store(store=store, path=path, table=dataset)
        self._raise_store_type_not_supported()

    def write_from_pandas(self, table: DataFrame) -> None:
        """Write the table from a pandas DataFrame."""
        table = normalize_pandas_df(
            table,
            index_key=self.index_key,
            index_type=self.index_type,
            reset_index=True,
        )
        table = pa.Table.from_pandas(table, preserve_index=False)
        self._write_pyarrow_dataset(table)

    def write_from_polars(self, table: PolarsDataFrame | LazyFrame) -> None:
        """Write the table from a polars DataFrame or LazyFrame."""
        table = normalize_polars_lf(
            table,
            index_key=self.index_key,
            index_type=self.index_type,
        )

        if isinstance(table, LazyFrame):
            table = table.collect()
        table = table.to_arrow()
        self._write_pyarrow_dataset(table)



from collections.abc import Sequence
from typing import Protocol

import dask.array as da
import numpy as np

from ngio.io_pipes._ops_axes import AxesOps
from ngio.io_pipes._ops_slices import SlicingOps


class TransformProtocol(Protocol):
    """Protocol for a generic transform."""

    def get_as_numpy_transform(
        self, array: np.ndarray, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> np.ndarray:
        """A transformation to be applied after loading a numpy array."""
        ...

    def get_as_dask_transform(
        self, array: da.Array, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> da.Array:
        """A transformation to be applied after loading a dask array."""
        ...

    def set_as_numpy_transform(
        self, array: np.ndarray, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> np.ndarray:
        """A transformation to be applied before writing a numpy array."""
        ...

    def set_as_dask_transform(
        self, array: da.Array, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> da.Array:
        """A transformation to be applied before writing a dask array."""
        ...


def get_as_numpy_transform(
    array: np.ndarray,
    slicing_ops: SlicingOps,
    axes_ops: AxesOps,
    transforms: Sequence[TransformProtocol] | None = None,
) -> np.ndarray:
    """Apply a numpy transform to an array."""
    if transforms is None:
        return array

    for transform in transforms:
        array = transform.get_as_numpy_transform(
            array, slicing_ops=slicing_ops, axes_ops=axes_ops
        )
    return array


def get_as_dask_transform(
    array: da.Array,
    slicing_ops: SlicingOps,
    axes_ops: AxesOps,
    transforms: Sequence[TransformProtocol] | None = None,
) -> da.Array:
    """Apply a dask transform to an array."""
    if transforms is None:
        return array

    for transform in transforms:
        array = transform.get_as_dask_transform(
            array, slicing_ops=slicing_ops, axes_ops=axes_ops
        )
    return array


def set_as_numpy_transform(
    array: np.ndarray,
    slicing_ops: SlicingOps,
    axes_ops: AxesOps,
    transforms: Sequence[TransformProtocol] | None = None,
) -> np.ndarray:
    """Apply inverse numpy transforms to an array."""
    if transforms is None:
        return array

    for transform in transforms:
        array = transform.set_as_numpy_transform(
            array, slicing_ops=slicing_ops, axes_ops=axes_ops
        )
    return array


def set_as_dask_transform(
    array: da.Array,
    slicing_ops: SlicingOps,
    axes_ops: AxesOps,
    transforms: Sequence[TransformProtocol] | None = None,
) -> da.Array:
    """Apply inverse dask transforms to an array."""
    if transforms is None:
        return array

    for transform in transforms:
        array = transform.set_as_dask_transform(
            array, slicing_ops=slicing_ops, axes_ops=axes_ops
        )
    return array
